"""Taegis Magic ipydatagrid widget."""

import json
import logging
from enum import Enum
from typing import Any, Dict, Optional

import pandas as pd
from ipydatagrid import DataGrid, HyperlinkRenderer, VegaExpr

log = logging.getLogger(__name__)


def validate_data_map(value: Any) -> Any:
    """Certain data types show as [object Object] in the datagrid.  This function will convert them to a string."""
    if isinstance(value, Enum):
        try:
            return value.value
        except Exception as e:
            log.error(f"Error converting Enum ({value}) to value: {e}")
            return "Error"

    if isinstance(value, (dict, list)):
        try:
            return json.dumps(value)
        except Exception as e:
            log.error(f"Error converting JSON-able ({value}) to value: {e}")
            return "Error"

    return value


def data_grid(
    df: pd.DataFrame,
    *,
    auto_fit_columns: bool = True,
    auto_fit_params: Optional[Dict[str, Any]] = None,
    validate_data: bool = True,
    editable: bool = True,
    limit: Optional[int] = 1000,
    **kwargs,
) -> DataGrid:
    """Default configurations for an ipydatagrid widget.

    Parameters
    ----------
    df : DataFrame
        Pandas DataFrame
    auto_fit_params : Optional[Dict[str, Any]], optional
        Data fit options, by default None
        https://github.com/jupyter-widgets/ipydatagrid/blob/main/examples/Column%20Width%20Auto-Fit.ipynb
    validate_data : bool, optional
        Change data that would present [object Object] to a string value, by default True
    editable : bool, optional
        Allow editing of the data in the widget, by default True
    limit: Optional[int], optional
        Limit the number of rows displayed in the widget, by default 1000
    kwargs: Dict[str, Any], optional
        Keyword arguments for the DataGrid widget

    Returns
    -------
    DataGrid
        ipydatagrid widget
    """
    df = df.copy()
    if limit:
        df = df.head(limit)

    if validate_data:
        # targeting pandas 2.0+, should update to just 'map' when we want to move to 2.1+
        log.debug("Validating dataframe...")
        df = df.applymap(validate_data_map, na_action="ignore")

    if "share_link" in df.columns:
        if "renderers" not in kwargs:
            log.debug("Renderers not found.  Adding default rendererer...")
            kwargs["renderers"] = {}
        if "share_link" not in kwargs["renderers"]:
            log.debug(
                "share_link renderer not found.  Adding default share_link rendererer..."
            )
            share_link_renderer = HyperlinkRenderer(
                url=VegaExpr("cell.value"),
                url_name=VegaExpr("cell.value"),
                text_color="blue",
            )
            kwargs["renderers"].update({"share_link": share_link_renderer})

    log.debug("Building widget...")
    grid = DataGrid(df, editable=editable, **kwargs)
    if auto_fit_columns:
        if not auto_fit_params:
            auto_fit_params = {"area": "body", "padding": 200, "numCols": None}

        grid.auto_fit_params = auto_fit_params
        grid.auto_fit_columns = auto_fit_columns

    log.debug("Returning widget...")
    return grid



"""Taegis Magic Service Generator."""

from taegis_magic._version import __version__
from taegis_magic.commands.configure import (
    AUTH_SECTION,
    REGIONS_SECTION,
    use_universal_auth,
)

from taegis_sdk_python import GraphQLService
from taegis_sdk_python._consts import TAEGIS_ENVIRONMENT_URLS
from taegis_sdk_python.config import get_config


def get_service(*args, **kwargs) -> GraphQLService:
    """Get a configured Taegis GraphQL Service object."""
    config = get_config()

    if not config.has_section(AUTH_SECTION):
        config.add_section(AUTH_SECTION)

    if not config.has_section(REGIONS_SECTION):
        config.add_section(REGIONS_SECTION)

    regions = config[REGIONS_SECTION]
    additional_regions = {k: v for k, v in regions.items() if k and v}

    environments = {**TAEGIS_ENVIRONMENT_URLS, **additional_regions}

    if (
        use_universal_authentication := kwargs.get("use_universal_authentication")
    ) is None:
        use_universal_authentication = config[AUTH_SECTION].getboolean(
            "use_universal_authentication", False
        )

    kwargs.pop("use_universal_authentication", None)

    if "environments" in kwargs:
        environments.update(kwargs["environments"])
        kwargs.pop("environments", None)

    extra_headers = {
        "User-Agent": f"taegis_magic/{__version__}",
        "apollographql-client-name": "taegis_magic",
        "apollographql-client-version": __version__,
    }

    if "extra_headers" in kwargs:
        extra_headers.update(kwargs["extra_headers"])
        kwargs.pop("extra_headers", None)

    return GraphQLService(
        environments=environments,
        extra_headers=extra_headers,
        use_universal_authentication=use_universal_authentication,
        *args,
        **kwargs,
    )



import inspect
import logging
from dataclasses import dataclass
from typing import List, Optional

import typer
from taegis_magic.commands.configure import AUTH_SECTION, REGIONS_SECTION
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResult
from taegis_magic.core.service import get_service
from typing_extensions import Annotated

from taegis_sdk_python._consts import TAEGIS_ENVIRONMENT_URLS
from taegis_sdk_python.config import get_config, write_config

log = logging.getLogger(__name__)

CONFIG = get_config()
if not CONFIG.has_section(AUTH_SECTION):
    CONFIG.add_section(AUTH_SECTION)

app = typer.Typer(help="Taegis Authentication Commands.")


@dataclass
class AuthenticationResult:
    action: str


@app.command()
@tracing
def login(
    region: Annotated[Optional[List[str]], typer.Option(help="Taegis Region")] = None,
    use_universal_authentication: Annotated[
        bool,
        typer.Option(
            help="Setup authentication to use unified authentication endpoint."
        ),
    ] = CONFIG[AUTH_SECTION].getboolean("use_universal_authentication", fallback=False),
):
    """Log into Taegis environments.  Use '--region all' authenticate each of Taegis environments."""
    service = get_service(use_universal_authentication=use_universal_authentication)

    config = get_config()
    if not config.has_section(REGIONS_SECTION):
        config.add_section(REGIONS_SECTION)

    if not config.has_section(AUTH_SECTION):
        config.add_section(AUTH_SECTION)

    config[AUTH_SECTION]["use_universal_authentication"] = str(
        use_universal_authentication
    ).lower()

    write_config(config)

    regions = config[REGIONS_SECTION]
    additional_regions = {k: v for k, v in regions.items() if k and v}

    environments = {**TAEGIS_ENVIRONMENT_URLS, **additional_regions}

    if service.use_universal_authentication:
        print(f"Login for region: {service.authentication_environment}")
        service.access_token
    else:
        if region:
            if "all" in region:
                region = []
                urls = []
                for r, url in environments.items():
                    if url in urls:
                        continue
                    urls.append(url)
                    region.append(r)
            for r in region:
                print(f"Login for region: {environments[r]}")
                with service(environment=r):
                    service.access_token
        else:
            print(f"Login for region: {service.environment}")
            service.access_token

    return TaegisResult(
        raw_results=AuthenticationResult(action="login"),
        service="authentication",
        tenant_id=None,
        region=region,
        arguments=inspect.currentframe().f_locals,
    )


@app.command()
@tracing
def logout(
    region: Annotated[Optional[List[str]], typer.Option(help="Taegis Region")] = None,
    use_universal_authentication: Annotated[
        bool,
        typer.Option(
            help="Setup authentication to use unified authentication endpoint."
        ),
    ] = CONFIG[AUTH_SECTION].getboolean("use_universal_authentication", fallback=False),
):
    """Logout of Taegis environments.  Use '--region all' fully logout of Taegis."""
    service = get_service(use_universal_authentication=use_universal_authentication)

    config = get_config()
    if not config.has_section(REGIONS_SECTION):
        config.add_section(REGIONS_SECTION)

    if not config.has_section(AUTH_SECTION):
        config.add_section(AUTH_SECTION)

    config[AUTH_SECTION]["use_universal_authentication"] = str(
        use_universal_authentication
    ).lower()

    write_config(config)

    regions = config[REGIONS_SECTION]
    additional_regions = {k: v for k, v in regions.items() if k and v}

    environments = {**TAEGIS_ENVIRONMENT_URLS, **additional_regions}

    if service.use_universal_authentication:
        print(f"Logout for region: {service.authentication_environment}")
        service.clear_access_token()
    else:
        if region:
            if "all" in region:
                region = []
                urls = []
                for r, url in environments.items():
                    if url in urls:
                        continue
                    urls.append(url)
                    region.append(r)
            for r in region:
                print(f"Logout for region: {environments[r]}")
                with service(environment=r):
                    service.clear_access_token()
        else:
            print(f"Logout for region: {service.environment}")
            service.clear_access_token()

    return TaegisResult(
        raw_results=AuthenticationResult(action="logout"),
        service="authentication",
        tenant_id=None,
        region=region,
        arguments=inspect.currentframe().f_locals,
    )



"""Taegis Magic tenant-profiles commands."""

import inspect
import logging
import warnings
from dataclasses import asdict, field
from enum import Enum
from pathlib import Path
from typing import Any, List, Optional

import pandas as pd
import numpy as np
import typer
from taegis_magic.core.callbacks import verify_file
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_sdk_python import GraphQLNoRowsInResultSetError
from taegis_sdk_python.services.tenant_profiles.types import (
    CriticalContactMtpInput,
    CustomerContactPreferenceMtp,
    MfaAccessCreateMtpInput,
    MfaAccessMtp,
    MfaAccessUpdateMtpInput,
    MfaServiceMtp,
    MtpNetworkType,
    NetworkRangeCreateMtpInput,
    NetworkRangeMtp,
    NetworkRangeUpdateMtpInput,
    SecurityControlCreateMtpInput,
    SecurityControlMtp,
    SecurityControlServiceMtp,
    SecurityControlSourceMtp,
    SecurityControlUpdateMtpInput,
)
from typing_extensions import Annotated

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Tenant Profile Commands.")
contacts = typer.Typer(help="Manage tenant profile contacts.")
network = typer.Typer(help="Manage tenant profile network ranges.")
network_template = typer.Typer(help="Manager tenant network range with template files.")
note = typer.Typer(help="Manage tenant profile notes.")
security_controls = typer.Typer(
    help="Manage tenant profile security control device information.",
)
mfa = typer.Typer(help="Manage tenant profile multifactor authentication information.")


network.add_typer(network_template, name="template")
app.add_typer(contacts, name="contacts")
app.add_typer(network, name="network")
app.add_typer(note, name="note")
app.add_typer(security_controls, name="security-controls")
app.add_typer(mfa, name="mfa")


def excel_normalize(value):
    if isinstance(value, bool):
        return str(value)
    if isinstance(value, Enum):
        return value.value
    return value


class TaegisTenantProfileResultNormalizer(TaegisResultsNormalizer):
    """Tenant Profiles single response normalizer."""

    raw_results: Any = field(default=None)

    @property
    def results(self):
        return [asdict(self.raw_results)]


class TaegisTenantProfileResultsNormalizer(TaegisResultsNormalizer):
    """Tenant Profiles response list normalizer."""

    raw_results: List[Any] = field(default_factory=list)

    @property
    def results(self):
        return [asdict(r) for r in self.raw_results]


@app.command(name="list")
@tracing
def tenant_profile_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.managed_tenant_profile()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@contacts.command(name="list")
@tracing
def contacts_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List contacts in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.cse_contacts_mtp()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@contacts.command(name="add")
@tracing
def contacts_add(
    user_id: Annotated[str, typer.Option()],
    preference: Annotated[CustomerContactPreferenceMtp, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Add a contact in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.create_critical_contact_mtp(
        CriticalContactMtpInput(
            tdr_user_id=user_id,
            preference=preference,
        )
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@contacts.command(name="update")
@tracing
def contacts_update(
    id_: Annotated[str, typer.Option("--id")],
    user_id: Annotated[Optional[str], typer.Option()] = None,
    preference: Annotated[
        Optional[CustomerContactPreferenceMtp], typer.Option()
    ] = None,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Update a contact in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.update_critical_contact_mtp(
        id_=id_,
        input_=CriticalContactMtpInput(
            tdr_user_id=user_id,
            preference=preference,
        ),
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@contacts.command(name="remove")
@tracing
def contacts_remove(
    id_: Annotated[Optional[str], typer.Option("--id")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Remove a contact in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.delete_critical_contact_mtp(id_=id_)

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network.command(name="list")
@tracing
def network_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List network ranges in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.network_ranges_mtp()

    normalized_results = TaegisTenantProfileResultsNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network.command(name="add")
@tracing
def network_add(
    cidr: Annotated[str, typer.Option()],
    description: Annotated[str, typer.Option()],
    network_type: Annotated[MtpNetworkType, typer.Option()],
    is_critical: Annotated[bool, typer.Option()] = False,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Add a network range to tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.create_network_range_mtp(
        NetworkRangeCreateMtpInput(
            cidr=cidr,
            description=description,
            is_critical=is_critical,
            network_type=network_type,
        )
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network.command(name="update")
@tracing
def network_update(
    id_: Annotated[str, typer.Option("--id")],
    cidr: Annotated[Optional[str], typer.Option()] = None,
    description: Annotated[Optional[str], typer.Option()] = None,
    network_type: Annotated[Optional[MtpNetworkType], typer.Option()] = None,
    is_critical: Annotated[Optional[bool], typer.Option()] = None,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Update a network range in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.update_network_range_mtp(
        id_=id_,
        network=NetworkRangeUpdateMtpInput(
            cidr=cidr,
            description=description,
            is_critical=is_critical,
            network_type=network_type,
        ),
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network.command(name="remove")
@tracing
def network_remove(
    id_: Annotated[Optional[str], typer.Option("--id")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Remove a network range in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        results = service.tenant_profiles.mutation.delete_network_range_mtp(id_=id_)
    except GraphQLNoRowsInResultSetError:
        results = NetworkRangeMtp()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@note.command(name="list")
@tracing
def note_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List the note in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.note_mtp()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@note.command(name="update")
@tracing
def note_update(
    contents: Annotated[str, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Update the note in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.update_note_mtp(contents=contents)

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@security_controls.command(name="list")
@tracing
def security_controls_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List security controls in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.security_controls_mtp()

    normalized_results = TaegisTenantProfileResultsNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@security_controls.command(name="add")
@tracing
def security_controls_add(
    ip: Annotated[str, typer.Option()],
    details: Annotated[str, typer.Option()],
    service_: Annotated[SecurityControlServiceMtp, typer.Option("--service")],
    source: Annotated[SecurityControlSourceMtp, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Add a security control in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.create_security_control_mtp(
        SecurityControlCreateMtpInput(
            ip=ip,
            details=details,
            service=service_,
            source=source,
        )
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@security_controls.command(name="update")
@tracing
def security_controls_update(
    id_: Annotated[str, typer.Option("--id")],
    ip: Annotated[str, typer.Option()],
    details: Annotated[str, typer.Option()],
    service_: Annotated[SecurityControlServiceMtp, typer.Option("--service")],
    source: Annotated[SecurityControlSourceMtp, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Update a security control in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.update_security_control_mtp(
        id_=id_,
        input_=SecurityControlUpdateMtpInput(
            ip=ip,
            details=details,
            service=service_,
            source=source,
        ),
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@security_controls.command(name="remove")
@tracing
def security_controls_remove(
    id_: Annotated[str, typer.Option("--id")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Remove a security control in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        results = service.tenant_profiles.mutation.delete_security_control_mtp(id_=id_)
    except GraphQLNoRowsInResultSetError:
        results = SecurityControlMtp()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@mfa.command(name="list")
@tracing
def mfa_list(
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List MFA access in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.mfa_accesses_mtp()

    normalized_results = TaegisTenantProfileResultsNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@mfa.command(name="add")
@tracing
def mfa_add(
    ip: Annotated[Optional[str], typer.Option()],
    exceptions: Annotated[Optional[str], typer.Option()],
    details: Annotated[Optional[str], typer.Option()],
    service_: Annotated[Optional[MfaServiceMtp], typer.Option("--service")],
    mfa_required: Annotated[Optional[bool], typer.Option()] = False,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Add a MFA access in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.create_mfa_access_mtp(
        MfaAccessCreateMtpInput(
            ip=ip,
            mfa_required=mfa_required,
            exceptions=exceptions,
            details=details,
            service=service_,
        )
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@mfa.command(name="update")
@tracing
def mfa_update(
    id_: Annotated[Optional[str], typer.Option("--id")],
    ip: Annotated[Optional[str], typer.Option()],
    exceptions: Annotated[Optional[str], typer.Option()],
    details: Annotated[Optional[str], typer.Option()],
    service_: Annotated[Optional[MfaServiceMtp], typer.Option("--service")],
    mfa_required: Annotated[Optional[bool], typer.Option()] = False,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Update a MFA access in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.mutation.update_mfa_access_mtp(
        id_=id_,
        input_=MfaAccessUpdateMtpInput(
            ip=ip,
            mfa_required=mfa_required,
            exceptions=exceptions,
            details=details,
            service=service_,
        ),
    )

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@mfa.command(name="remove")
@tracing
def mfa_remove(
    id_: Annotated[Optional[str], typer.Option("--id")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Remove a MFA access in a tenant profile."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        results = service.tenant_profiles.mutation.delete_mfa_access_mtp(
            id_=id_,
        )
    except GraphQLNoRowsInResultSetError:
        results = MfaAccessMtp()

    normalized_results = TaegisTenantProfileResultNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network_template.command("generate")
@tracing
def network_template_generate(
    filename: Annotated[
        Path,
        typer.Option(
            exists=False,
            file_okay=True,
            dir_okay=False,
            readable=False,
            writable=True,
            resolve_path=True,
            callback=verify_file,
        ),
    ] = Path("taegis_network_range_template.xlsx"),
):
    """Generate a template file for tenant profiles network ranges."""

    input_ = NetworkRangeCreateMtpInput(
        cidr="192.0.2.0/24",
        description="RFC 5737 TEST-NET-1",
        network_type=MtpNetworkType.OTHER,
        is_critical=False,
    )
    df = pd.json_normalize([asdict(input_)], max_level=3)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df = df.applymap(excel_normalize, na_action="ignore")

    df.to_excel(filename, index=False)

    normalized_results = TaegisResultsNormalizer(
        raw_results=[{"filename": str(filename.resolve())}],
        service="tenant_profiles",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network_template.command("export")
@tracing
def network_template_export(
    filename: Annotated[
        Path,
        typer.Option(
            exists=False,
            file_okay=True,
            dir_okay=False,
            readable=False,
            writable=True,
            resolve_path=True,
            callback=verify_file,
        ),
    ] = Path("taegis_network_range_export.xlsx"),
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Export tenant profiles network ranges to a file."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.tenant_profiles.query.network_ranges_mtp()
    df = pd.json_normalize([asdict(result) for result in results], max_level=3)

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        df = df.applymap(excel_normalize, na_action="ignore")

    df.to_excel(filename, index=False)

    normalized_results = TaegisResultsNormalizer(
        raw_results=[{"filename": str(filename.resolve())}],
        service="tenant_profiles",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@network_template.command("upload")
@tracing
def network_template_upload(
    filename: Annotated[
        Path,
        typer.Option(
            exists=True,
            file_okay=True,
            dir_okay=False,
            readable=True,
            writable=False,
            resolve_path=True,
        ),
    ] = Path("taegis_network_range_template.xlsx"),
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Create/Update Tenant Profile Network Ranges."""
    service = get_service(environment=region, tenant_id=tenant)

    try:
        df = pd.read_excel(filename)
    except ValueError as exc:
        log.error(f"{filename.resolve()} is not an Excel file.")
        raise typer.Exit() from exc
    except FileNotFoundError as exc:
        log.error(f"{filename.resolve()} not found.")
        raise typer.Exit() from exc

    def excel_decode(row, networks):
        try:
            for network in networks:
                if network.cidr == row.cidr:
                    return NetworkRangeUpdateMtpInput(
                        cidr=row["cidr"],
                        description=row["description"],
                        is_critical=row["is_critical"],
                        network_type=MtpNetworkType(row["network_type"]),
                    )

            return NetworkRangeCreateMtpInput(
                cidr=row["cidr"],
                description=row["description"],
                is_critical=row["is_critical"],
                network_type=MtpNetworkType(row["network_type"]),
            )
        except ValueError as exc:
            if "MtpNetworkType" in exc.args[0]:
                log.error(
                    f"{exc}.  Use {[item.value for item in list(MtpNetworkType)]} instead."
                )
            else:
                log.error(exc)

            return np.nan
        except Exception as exc:
            log.error(exc)

            return np.nan

    networks = service.tenant_profiles.query.network_ranges_mtp()

    series = df.apply(excel_decode, axis=1, networks=networks).dropna()

    create_networks = [
        row for row in series if isinstance(row, NetworkRangeCreateMtpInput)
    ]
    update_networks = [
        row for row in series if isinstance(row, NetworkRangeUpdateMtpInput)
    ]

    for network in create_networks:
        try:
            results = service.tenant_profiles.mutation.create_network_range_mtp(network)
        except Exception as exc:
            log.error(exc)
            continue

    for network in update_networks:
        try:
            id_ = next(iter([n.id for n in networks if n.cidr == network.cidr]))
        except StopIteration:
            log.error(f"Could not find `id` for network: {network.cidr}")
            continue

        try:
            service.tenant_profiles.mutation.update_network_range_mtp(id_, network)
        except Exception as exc:
            log.error(exc)
            continue

    results = service.tenant_profiles.query.network_ranges_mtp()

    normalized_results = TaegisTenantProfileResultsNormalizer(
        raw_results=results,
        service="tenant_profiles",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results



"""Taegis Magic application logging."""

import logging
import functools
from typing import Callable


TRACE_LOG_LEVEL = 5
logging.addLevelName(TRACE_LOG_LEVEL, "TRACE")


def trace(self: logging.Logger, message: str, *args, **kwargs):
    """
    Trace log level for function decorator.

    Parameters
    ----------
    self : logging.Logger
        Logger
    message : _type_
        Message to log
    """
    if self.isEnabledFor(TRACE_LOG_LEVEL):
        self._log(TRACE_LOG_LEVEL, message, args, **kwargs)


logging.Logger.trace = trace

log = logging.getLogger(__name__)


def get_module_logger() -> logging.Logger:
    """
    Get the module logger object.

    Returns
    -------
    logging.Logger
        Taegis Magic module logger
    """
    logger = logging.getLogger("taegis_magic")
    logger.propagate = False
    logger.handlers = []

    handler = logging.StreamHandler()
    formatter = logging.Formatter(
        "%(asctime)s::%(levelname)s::%(name)s::%(message)s", datefmt="%Y-%m-%dT%H:%M:%S"
    )
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    return logger


def tracing(func: Callable):
    """
    Log entry and exit from function.

    Parameters
    ----------
    func : Callable
        Function to log
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        log.trace(f"Entering {func.__name__}(args: {args}, kwargs: {kwargs})...")

        try:
            result = func(*args, **kwargs)
        except Exception as exc:
            log.error(f"Exception raised in {func.__name__}. exception: {str(exc)}")
            raise exc

        log.trace(f"Exiting {func.__name__}...")

        return result

    return wrapper



"""Taegis Magic Sharelinks Commands."""

import logging
from typing import Optional
from urllib.parse import urlparse

import typer
from taegis_magic.commands.alerts import AlertsResultsNormalizer
from taegis_magic.commands.events import TaegisEventNormalizer
from taegis_magic.commands.investigations import InvestigationsCreatedResultsNormalizer
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_sdk_python.services.alerts.types import GetByIDRequestInput
from taegis_sdk_python.services.investigations2.types import InvestigationV2Arguments
from typing_extensions import Annotated

app = typer.Typer(help="Taegis Sharelinks Commands.")

log = logging.getLogger(__name__)


@app.command()
@tracing
def unfurl(
    id_: Annotated[str, typer.Option("--id", help="Taegis Sharelink URL or UUID")],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
) -> TaegisResultsNormalizer:
    """Unpack sharelink urls and uuids.  Returns underlying datastructure results."""
    if "/share/" in id_:
        parse_result = urlparse(id_)
        id_ = parse_result.path.replace("/share/", "")

    service = get_service(environment=region, tenant_id=tenant)

    results = service.sharelinks.query.share_link_by_id(id_=id_)

    if results.link_type == "alertId":
        with service(tenant_id=results.tenant_id):
            unfurl_results = service.alerts.query.alerts_service_retrieve_alerts_by_id(
                GetByIDRequestInput(i_ds=[results.link_ref])
            )

        normalized_results = AlertsResultsNormalizer(
            raw_results=[unfurl_results],
            service="alerts",
            tenant_id=results.tenant_id,
            region=service.environment,
            arguments={
                "id": id_,
                "region": service.environment,
                "tenant": service.tenant_id,
            },
        )
        normalized_results._shareable_url = (
            f'{service.core.sync_url.replace("api.", "")}/share/{id_}'
        )
    elif results.link_type == "eventId":
        with service(tenant_id=results.tenant_id):
            unfurl_results = service.events.query.events(ids=[results.link_ref])
            normalized_results = TaegisEventNormalizer(
                raw_results=unfurl_results,
                service="events",
                tenant_id=service.tenant_id,
                region=service.environment,
                arguments={
                    "id": id_,
                    "region": service.environment,
                    "tenant": service.tenant_id,
                },
            )
    elif results.link_type == "investigationId":
        with service(tenant_id=results.tenant_id):
            unfurl_results = service.investigations2.query.investigation_v2(
                InvestigationV2Arguments(id=results.link_ref)
            )
            normalized_results = InvestigationsCreatedResultsNormalizer(
                raw_results=unfurl_results,
                service="cases",
                tenant_id=service.tenant_id,
                region=service.environment,
                arguments={
                    "id": id_,
                    "region": service.environment,
                    "tenant": service.tenant_id,
                },
            )
    else:
        raise ValueError(f"Unable to process sharelink results: {results}")

    return normalized_results



"""Taegis Magic events commands."""

import inspect
import logging
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Optional

import typer
from dataclasses_json import dataclass_json
from taegis_magic.commands.configure import QUERIES_SECTION
from taegis_magic.commands.utils.investigations import insert_search_query
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResults, TaegisResultsNormalizer
from taegis_magic.core.service import get_service
from taegis_sdk_python.config import get_config
from taegis_sdk_python.services.events.types import (
    Event,
    EventQueryOptions,
    EventQueryResults,
)
from taegis_sdk_python.services.rules.types import RuleEventType
from taegis_sdk_python.services.sharelinks.types import (
    ExtraParamCreateInput,
    ShareLinkCreateInput,
)
from typing_extensions import Annotated

log = logging.getLogger(__name__)

CONFIG = get_config()
if not CONFIG.has_section(QUERIES_SECTION):
    CONFIG.add_section(QUERIES_SECTION)


app = typer.Typer(help="Taegis Events Commands.")


@dataclass_json
@dataclass
class TaegisEventNormalizer(TaegisResultsNormalizer):
    """Taegis Events Normalizer."""

    raw_results: List[Event] = field(default_factory=list)

    @property
    def results(self) -> List[Dict[str, Any]]:
        """Results from event by ids query.

        Returns
        -------
        List[Dict[str, Any]]
            List of Events
        """
        return [asdict(event) for event in self.raw_results]


@dataclass_json
@dataclass
class TaegisEventQueryNormalizer(TaegisResultsNormalizer):
    """Taegis Event Query Result Normalizer."""

    raw_results: List[EventQueryResults] = field(default_factory=list)
    query: str = ""
    is_saved: bool = False
    _query_id = None
    _shareable_url = None

    def _repr_markdown_(self):
        """Represent as markdown."""
        return self._display_template(
            "taegis_search_results.md.jinja"
        )  # pragma: no test

    @property
    def results(self) -> List[Dict[str, Any]]:
        """Results from Taegis event query.

        Returns
        -------
        List[Dict[str, Any]]
            List of Taegis event objects from query results.
        """

        rows = []
        for result in self.raw_results or []:
            if result.result:
                if result.result.rows:
                    rows.extend(result.result.rows)
                else:
                    log.info(f"No rows found in result: {result.result}")
            else:
                log.info(f"Non valid result set found: {result}")

        return rows

    @property
    def status(self) -> str:
        """Response status of Taegis event query.

        Returns
        -------
        str
            Command result status.
        """
        return self.raw_results[0].result.status if self.raw_results else "Error"

    @property
    def total_results(self) -> int:
        """Query results total number of results on server.

        This query type does not have a valid server return valid.

        Returns
        -------
        int
            Returns number of results.
        """
        # We cannot trust the result totals provided by the server.
        return -1

    @property
    def results_returned(self) -> int:
        """Query results total number of results returned.

        Returns
        -------
        str
            Returns number of results.
        """
        return len(self.results)

    @property
    def query_identifier(self) -> str:
        """Generate a query identifier for Taegis XDR QL Event queries.

        Returns
        -------
        str
            Query Identifier

        Raises
        ------
        ValueError
            No query found to generate query id
        ValueError
            No query id returned from Query API
        """
        if not self.raw_results:
            return None

        for result in self.raw_results:
            if result.query_id:
                return result.query_id

        return None

    @property
    def shareable_url(self) -> str:
        """Generate a shareable url for Taegis XDR.

        Returns
        -------
        str
            Share Link url.
        """
        if not self.raw_results:
            return "Not able to create shareable link"

        if not self.query_identifier:
            return "Unable to create shareable link"

        if self._shareable_url:
            return self._shareable_url

        service = get_service(environment=self.region, tenant_id=self.tenant_id)

        result = service.sharelinks.mutation.create_share_link(
            ShareLinkCreateInput(
                link_ref=self.query_identifier,
                link_target="cql",
                link_type="queryId",
                tenant_id=self.tenant_id,
                extra_parameters=[
                    ExtraParamCreateInput(key="sourceType", value="event"),
                ],
            )
        )

        self._shareable_url = (
            f'{service.core.sync_url.replace("api.", "")}/share/{result.id}'
        )
        return self._shareable_url


def get_next_page(events_results: List[EventQueryResults]) -> Optional[str]:
    """Retrieve events  next page indicator."""
    try:
        # the next page could be found in any of the result pages,
        # but we cannot garuntee which result it will be found in
        return next(
            iter({result.next for result in events_results if result.next is not None})
        )
    except StopIteration:
        return None


@app.command()
@tracing
def search(
    cell: Optional[str] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
    save: bool = False,
    track: Annotated[bool, typer.Option()] = CONFIG[QUERIES_SECTION].getboolean(
        "track",
        fallback=False,
    ),
    database: Annotated[str, typer.Option()] = ":memory:",
):
    """Taegis Events search."""
    if not cell:
        cell = ""

    service = get_service(tenant_id=tenant, environment=region)
    options = EventQueryOptions(
        timestamp_ascending=True,
        page_size=1000,
        max_rows=100000,
        aggregation_off=False,
    )
    results = []

    result = service.events.subscription.event_query(
        cell,
        options=options,
        metadata={
            "callerName": CONFIG[QUERIES_SECTION].get(
                "callername", fallback="Taegis Magic"
            ),
        },
    )
    results.extend(result)
    next_page = get_next_page(result)

    while next_page:
        result = service.events.subscription.event_page(next_page)
        results.extend(result)
        next_page = get_next_page(result)

    normalized_results = TaegisEventQueryNormalizer(
        raw_results=results,
        service="events",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments={
            "cell": cell,
            "tenant": service.tenant_id,
            "region": service.environment,
            "save": save,
            "track": track,
            "database": database,
        },
        query=cell,
        is_saved=save,
    )

    if track:
        insert_search_query(database, normalized_results)

    return normalized_results


@app.command()
@tracing
def events(
    resource_id: Annotated[Optional[List[str]], typer.Option()] = None,
    tenant: Optional[str] = None,
    region: Optional[str] = None,
):
    """Get events by resource id."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.events.query.events(ids=resource_id)

    normalized_results = TaegisEventNormalizer(
        raw_results=results,
        service="events",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command()
@tracing
def schema(
    type_: Annotated[RuleEventType, typer.Option("--type", help="Event Schema Type")],
    tenant: Annotated[Optional[str], typer.Option(help="Tenant ID")] = None,
    region: Annotated[Optional[str], typer.Option(help="Taegis Region")] = None,
):
    """Query Event Schema fields."""

    service = get_service(environment=region, tenant_id=tenant)

    schema_fields = service.rules.query.filter_keys(RuleEventType(type_))

    @dataclass_json
    @dataclass
    class SchemaField:
        key: str

    schema_fields = [SchemaField(f) for f in schema_fields]

    return TaegisResults(
        raw_results=schema_fields,
        service="events",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments={"type": type_, "tenant": tenant, "region": region},
    )



from typing import Any, Dict, List, Optional

import pandas as pd

from taegis_sdk_python import GraphQLService


def get_tenant_id_column(df: pd.DataFrame) -> str:
    tenant_column = None
    if "tenant_id" in df.columns:
        tenant_column = "tenant_id"
    elif "tenant.id" in df.columns:
        tenant_column = "tenant.id"
    elif "event_data.tenant_id" in df.columns:
        tenant_column = "event_data.tenant_id"

    if tenant_column is None:
        raise ValueError("Tenant ID column not found in DataFrame")

    return tenant_column


def get_first_tenant_id(service: GraphQLService) -> str:
    """Get the first tenant ID from the service's current subject.

    Parameters
    ----------
    service : GraphQLService
        The GraphQL service instance.

    Returns
    -------
    str
        The first tenant ID.
    """
    subject = service.subjects.query.current_subject()

    if not subject.role_assignment_data.assigned_tenant_ids:
        raise ValueError(
            "No tenant ID found in the current subject's role assignment data."
        )

    first_tenant_id = subject.role_assignment_data.assigned_tenant_ids[0]

    return first_tenant_id


def remove_output_node(
    output: str, node: str, start: Optional[int] = None, end: Optional[int] = None
) -> str:
    """Remove a GraphQL Output node from the output string.

    Parameters
    ----------
    output : str
        GraphQL output string
    node : str
        Node identifier
    start : Optional[int], optional
        index to start node search useful for nodes sharing a name, by default None
    end : Optional[int], optional
        index to end node search useful for nodes sharing a name, by default None

    Returns
    -------
    str
        Modified output string
    """
    try:
        start_idx = output.index(f" {node} ", start, end) + 1
    except ValueError:
        return output

    end_idx = None

    brackets_found = 0
    for idx, char in enumerate(output):
        if idx < start_idx + len(node):
            continue

        if char == " ":
            continue

        if brackets_found == 0 and char != "{":
            end_idx = idx
            break

        if char == "{":
            brackets_found += 1

        elif char == "}":
            brackets_found -= 1

        end_idx = idx + 1

    return output[:start_idx] + output[end_idx:]


def to_dataframe(results: List[Dict[str, Any]]) -> pd.DataFrame:
    """Ingest a list of results and convert it to a DataFrame that contains no blank columns.

    Parameters
    ----------
    results : List[Dict[str, Any]]
        A list of dictionary results

    Returns
    -------
    pd.DataFrame
        Returns a dataFrame with no blank columns.
    """

    return pd.json_normalize(results, max_level=3).dropna(how="all", axis=1)



"""Context gathering functions."""

from typing import Callable, Dict, List, Optional

import pandas as pd
import numpy as np
import panel as pn
from IPython.core.display import display
from taegis_magic.pandas.alerts import inflate_raw_events
from taegis_sdk_python import GraphQLService

pn.extension("tabulator")


def normalize_entities(df: pd.DataFrame) -> pd.DataFrame:
    """Normalize entity values from alerts for Taegis Query Language logical types:

        '@ip', '@domain', '@hash', '@host', '@user'.

    https://docs.ctpx.secureworks.com/search/querylanguage/advanced_search/#logical-types

    Parameters
    ----------
    df : pd.DataFrame
        Alert DataFrame

    Returns
    -------
    pd.DataFrame
        Normalized Entity DataFrame
    """
    df = df.copy()

    df = df.explode("entities.entities")

    def normalize_entity(entity):
        entity_type, value = entity.split(":", 1)

        logical_types_map = {
            "@ip": [
                "ipaddress",
                "sourceIpAddress",
                "ipAddress",
                "destIpAddress",
                "sourceIpGeo",
                "destIpGeo",
                "receiverIp",
                "senderIp",
                "sourceAddress",
                "targetIp",
                "targetIpAddress",
                "destAddress",
                "destinationAddress",
            ],
            "@domain": [
                "dnsName",
                "domainname",
                "ipDomain",
                "topPrivateIpDomain",
                "domainName",
                "queryName",
                "sourceHostnameFqdn",
                "targetHostnameFqdn",
                "uriHost",
                "domain",
            ],
            "@hash": ["md5", "sha1", "sha256", "sha512"],
            "@host": [
                "sourceHostName",
                "destHostName",
                "workstationName",
                "targetHostName",
                "hostName",
                "computerName",
            ],
            "@user": ["userName", "username", "sourceUserName", "targetUserName"],
        }

        for logical_type_key, logical_type_values in logical_types_map.items():
            if entity_type in logical_type_values:
                return pd.Series(
                    [logical_type_key, value],
                    index=[
                        "taegis_magic.entities.field",
                        "taegis_magic.entities.value",
                    ],
                )

        return pd.Series(
            [None, None],
            index=["taegis_magic.entities.field", "taegis_magic.entities.value"],
        )

    df2 = (
        pd.concat([df, df["entities.entities"].apply(normalize_entity)], axis=1)
        .reset_index(drop=True)
        .dropna(
            subset=["taegis_magic.entities.field", "taegis_magic.entities.value"],
            how="all",
        )
        .drop_duplicates(
            subset=["id", "taegis_magic.entities.field", "taegis_magic.entities.value"]
        )
    )

    return df2


def relate_entities(df: pd.DataFrame) -> pd.DataFrame:
    """Relate logical entities to lists of indicators.

    Parameters
    ----------
    df : pd.DataFrame
        Normalized Entity DataFrame

    Returns
    -------
    pd.DataFrame
        Releted Entity DataFrame

    Raises
    ------
    ValueError
        column not found, run pipe function 'normalize_entities'
    """
    if df.empty:
        return df

    if "taegis_magic.entities.field" not in df.columns:
        raise ValueError(
            "taegis_magic.entities.field column not found, run pipe function 'normalize_entities'"
        )

    if "taegis_magic.entities.value" not in df.columns:
        raise ValueError(
            "taegis_magic.entities.value column not found, run pipe function 'normalize_entities'"
        )

    df = df.copy()

    groupby_alerts = (
        df.groupby(["taegis_magic.entities.field", "taegis_magic.entities.value"])["id"]
        .unique()
        .reset_index()
    )
    groupby_size = (
        df.groupby(["id", "taegis_magic.entities.field", "taegis_magic.entities.value"])
        .size()
        .reset_index()
    )

    def entity_relationships(entity: str, field: str):
        return (
            groupby_size[
                (
                    groupby_size["id"].isin(
                        groupby_alerts[
                            groupby_alerts["taegis_magic.entities.value"] == entity
                        ]["id"]
                        .explode()
                        .unique()
                    )
                )
                & (groupby_size["taegis_magic.entities.value"] != entity)
                & (groupby_size["taegis_magic.entities.field"] == field)
            ]["taegis_magic.entities.value"]
            .unique()
            .tolist()
        )

    for field in df["taegis_magic.entities.field"].unique():
        df[field] = df["taegis_magic.entities.value"].apply(
            entity_relationships, field=field
        )

    return df.drop_duplicates("taegis_magic.entities.value")


def generate_context_queries(
    df: pd.DataFrame,
    open_alerts_timeframe="EARLIEST=-1d",
    resolved_alerts_timeframe="EARLIEST=-30d",
    investigations_timeframe="EARLIEST=-30d",
    events_timeframe="EARLIEST=-1d",
) -> pd.DataFrame:
    """Generate Taegis Query Language queries based on logical entities.

    Parameters
    ----------
    df : pd.DataFrame
        Normalized Entities DataFrame

    Returns
    -------
    pd.DataFrame
        Generate Query DataFrame
    """

    if df.empty:
        return df

    if "taegis_magic.entities.field" not in df.columns:
        raise ValueError(
            "taegis_magic.entities.field column not found, run pipe function 'normalize_entities'"
        )

    if "taegis_magic.entities.value" not in df.columns:
        raise ValueError(
            "taegis_magic.entities.value column not found, run pipe function 'normalize_entities'"
        )

    df = df.copy()

    def generate_context_from_row(row):
        query_params = [
            f"{row['taegis_magic.entities.field']} = '{row['taegis_magic.entities.value']}'"
        ]

        title = row["metadata.title"]

        for idx in row.index:
            if "@" in idx and row[idx]:
                for item in row[idx]:
                    query_params.append(f"{idx} = '{item}'")

        open_alerts_query = f"""
        FROM alert
        WHERE
            ({' OR '.join(query_params)}) AND
            metadata.title != '{title}' AND
            status = 'OPEN' AND
            investigation_ids IS NULL
        {open_alerts_timeframe}
        """

        resolved_alerts_query = f"""
        FROM alert
        WHERE
            ({' OR '.join(query_params)}) AND
            status != 'OPEN'
        {resolved_alerts_timeframe} | aggregate count by metadata.title, entities, status, resolution_reason
        """

        investigations_query = f"""
        FROM alert
        WHERE
            ({' OR '.join(query_params)}) AND
            investigation_ids IS NOT NULL
        {investigations_timeframe} | aggregate count by metadata.title, entities, investigation_ids, status
        """

        events_query = f"""
        WHERE
            ({' OR '.join(query_params)})
        {events_timeframe}
        """

        return pd.Series(
            [
                open_alerts_query.strip(),
                resolved_alerts_query.strip(),
                investigations_query.strip(),
                events_query.strip(),
            ],
            index=[
                "taegis_magic.open_alerts_query",
                "taegis_magic.resolved_alerts_query",
                "taegis_magic.investigations_query",
                "taegis_magic.events_query",
            ],
        )

    df2 = pd.concat(
        [df, df.apply(generate_context_from_row, axis=1)], axis=1
    ).reset_index(drop=True)

    return df2


def get_facet(df: pd.DataFrame, columns: List[str], title: str) -> pn.Card:
    for _ in columns.copy():
        try:
            df = (
                df.groupby(columns)
                .size()
                .rename("count")
                .sort_values(ascending=False)
                .to_frame()
                .reset_index()
            )
        except (KeyError, TypeError) as exc:
            columns.remove(exc.args[0])
            continue

        break

    header_filters = {
        column: {"type": "input", "func": "like", "placeholder": "Filter"}
        for column in df.columns
    }

    widget = pn.widgets.Tabulator(
        df,
        width=1050,
        height=500,
        formatters={"bool": {"type": "tickCross"}},
        layout="fit_data",
        theme="bootstrap5",
        theme_classes=["thead-dark", "table-sm"],
        # pagination="local",
        # page_size=50,
        header_filters=header_filters,
    )
    card = pn.Card(widget, title=title)
    return card


def display_facets(
    queries: Dict[str, pd.DataFrame], additional_columns: Optional[List[str]] = None
):
    """Display Card facets for each entity.

    Parameters
    ----------
    queries : Dict[str, pd.DataFrame]
        The DataFrame results from the entity queries.
    additional_columns : Optional[List[str]], optional
        Additional columns to display on each facet (usually provided by add_threat_intel), by default None
    """
    for entity in queries:
        if not queries[entity]["open_alerts"].empty:
            open_alerts = get_facet(
                df=queries[entity]["open_alerts"].explode("entities.entities"),
                columns=[
                    "metadata.title",
                    "entities.entities",
                ],
                title="Open Alerts",
            )
        else:
            open_alerts = pn.Card(title="Open Alerts")

        if not queries[entity]["investigations"].empty:
            investigations = get_facet(
                queries[entity]["investigations"],
                columns=[
                    "metadata.title",
                    "entities",
                    "investigation_ids",
                    "status",
                ],
                title="Investigations",
            )
        else:
            investigations = pn.Card(title="Investigations")

        if not queries[entity]["resolved_alerts"].empty:
            resolved_alerts = get_facet(
                df=queries[entity]["resolved_alerts"],
                columns=["metadata.title", "entities", "status", "resolution_reason"],
                title="Resolved Alerts",
            )
        else:
            resolved_alerts = pn.Card(title="Resolved Alerts")

        schema_cards = []

        try:
            schemas = (
                queries[entity]["events"]["resource_id"]
                .apply(lambda x: x.split(":")[2].split(".")[1])
                .unique()
            )
        except Exception:
            schemas = []

        for schema in schemas:
            df = queries[entity]["events"][
                queries[entity]["events"]["resource_id"].str.contains(schema)
            ].dropna(how="all", axis=1, inplace=False)
            if schema == "auth":
                columns = [
                    "source_address",
                    "target_user_name",
                    "action",
                    "auth_system",
                    "user_agent",
                    "application_name",
                ]
            elif schema == "thirdpartyalert":
                columns = [
                    "source_address",
                    "user_principal_name",
                    "title",
                    "ontology",
                ]
            elif schema == "dnsquery":
                columns = [
                    "hostname",
                    "os.os",
                    "os.arch",
                    "processcorrelationid.pid",
                    "query_name",
                    "query_type",
                ]
            elif schema == "cloudaudit":
                columns = [
                    "source_address",
                    "user_name",
                    "event_type",
                    "event_name",
                    "mfa_used",
                    "user_agent",
                ]
            elif schema == "http":
                columns = [
                    "source_username",
                    "source_address",
                    "destination_address",
                    "destination_port",
                    "user_agent",
                    "http_method",
                    "response_code",
                    "uri_host",
                    "uri_path",
                    "tx_byte_count",
                    "rx_byte_count",
                ]
            elif schema == "netflow":
                columns = [
                    "hostname",
                    "sensor_type",
                    "protocol",
                    "source_address",
                    "destination_address",
                    "destination_port",
                    "dns_name",
                ]
            elif schema == "nids":
                columns = [
                    "source_address",
                    "destination_address",
                    "action",
                    "blocked",
                    "enrichSummary",
                ]
            elif schema == "process":
                columns = [
                    "hostname",
                    "os.os",
                    "os.arch",
                    "sensor_type",
                    "username",
                    "user_is_admin",
                    "process_is_admin",
                    "image_path",
                    "commandline",
                    "was_blocked",
                ]
            elif schema == "scriptblock":
                df["decoded_block_text_truncated"] = df["decoded_block_text"].apply(
                    lambda x: x[:200]
                )
                columns = [
                    "os.os",
                    "os.arch",
                    "interpreter_name",
                    "interpreter_path",
                    "script_name",
                    "decoded_block_text_truncated",
                ]
            else:
                columns = list(df.columns)

            if additional_columns:
                columns.extend(additional_columns)

            schema_cards.append(get_facet(df=df, columns=columns, title=schema))

        display(
            pn.Card(
                open_alerts,
                investigations,
                resolved_alerts,
                *schema_cards,
                title=entity,
            )
        )


def add_threat_intel(
    df: pd.DataFrame,
    correlations: List[Callable],
    tenant_id: Optional[str] = None,
    region: Optional[str] = None,
) -> pd.DataFrame:
    """Correlate Threat Indicators to logical types.

    Parameters
    ----------
    df : pd.DataFrame
        Alerts/Events DataFrame.
    correlations : List[Callable]
        List of correlation functions.
    tenant_id : Optional[str]
        Teanant ID.
    region : Optional[str]
        Taegis Region.

    Returns
    -------
    pd.DataFrame
        Correlated DataFrame.

    Raises
    ------
    ValueError
        resource_id or event_data.resource_id not found in DataFrame.
    """
    if df.empty:
        return df

    df = df.copy()

    column = ""
    prefix = ""
    if "event_ids" in df.columns:
        df = df.pipe(inflate_raw_events)
        prefix = "event_data."

    column = f"{prefix}resource_id"

    if column not in df.columns:
        return df

    # https://docs.ctpx.secureworks.com/search/builder/advanced_search/#logical-type-mappings
    # we only need to define for @ip, @domain, and @hash
    ip_field_map = {
        "auth": [
            "target_address",
            "source_address",
        ],
        "cloudaudit": [
            "source_address",
        ],
        "dnsquery": [
            "source_address",
            "destination_address",
        ],
        "http": [
            "source_address",
            "destination_address",
            "true_source_address",
        ],
        "netflow": [
            "source_address",
            "destination_address",
            "source_nat_address",
            "destination_nat_address",
        ],
        "nids": [
            "source_address",
            "destination_address",
        ],
    }

    domain_field_map = {
        "auth": [
            "target_domain_name",
            "source_domain_name",
            "extra_targetoutbounddomainname",
        ],
        "dnsquery": [
            "query_name",
        ],
    }

    hash_field_map = {
        "auth": [
            "process_file_hash",
            "process_file_hash.md5",
            "process_file_hash.sha1",
            "process_file_hash.sha256",
            "process_file_hash.sha512",
        ],
        "filemod": [
            "file_hash,parent_process_file_hash.md5",
            "parent_process_file_hash.sha1",
            "parent_process_file_hash.sha256",
            "parent_process_file_hash.sha512",
            "process_file_hash.md5",
            "process_file_hash.sha1",
            "process_file_hash.sha256",
            "process_file_hash.sha512",
            "file_hash.md5",
            "file_hash.sha1",
            "file_hash.sha256",
            "file_hash.sha512",
        ],
        "process": [
            "program_hash.md5",
            "program_hash.sha1",
            "program_hash.sha256",
            "program_hash.sha512",
            "target_program.sha1_hash",
            "host_program.sha1_hash",
        ],
    }

    def parse_fields(row, field_map, prefix: str = ""):
        for schema, field_list in field_map.items():
            if schema in row[f"{prefix}resource_id"]:
                l = list(
                    set(
                        [
                            row[f"{prefix}{field}"]
                            for field in field_list
                            if f"{prefix}{field}" in row
                        ]
                    )
                )
                if np.nan in l:
                    l.remove(np.nan)
                return l
        return []

    df["logicals.ip"] = df.apply(
        parse_fields, field_map=ip_field_map, prefix=prefix, axis=1
    )
    df["logicals.domain"] = df.apply(
        parse_fields, field_map=domain_field_map, prefix=prefix, axis=1
    )
    df["logicals.hash"] = df.apply(
        parse_fields, field_map=hash_field_map, prefix=prefix, axis=1
    )

    logical_ips = set()
    logical_domains = set()
    logical_hashes = set()

    for _, row in df.iterrows():
        logical_ips.update(row["logicals.ip"])
        logical_domains.update(row["logicals.domain"])
        logical_hashes.update(row["logicals.hash"])

    indicators = logical_ips | logical_domains | logical_hashes

    for callable in correlations:
        df = df.pipe(
            callable, indicators=indicators, tenant_id=tenant_id, region=region
        )

    return df


def get_ti_pubs(
    df: pd.DataFrame,
    indicators: List[str],
    tenant_id: Optional[str] = None,
    region: Optional[str] = None,
) -> pd.DataFrame:
    """Correlate Threat Indicators to Threat Intelligence Publications.

    Parameters
    ----------
    df : pd.DataFrame
        Alerts/Events DataFrame.
    indicators : List[str]
        List of indicators; IPs, Domains, Hashes.
    tenant_id : Optional[str], optional
        Taegis Tenant ID, by default None
    region : Optional[str], optional
        Taegis Region, by default None

    Returns
    -------
    DataFrame
        Correlated DataFrame.
    """
    df = df.copy()

    service = GraphQLService(tenant_id=tenant_id, environment=region)

    # get TI Pubs
    ti_pubs = {}
    for indicator in indicators:
        with service(output="id Type Name"):
            results = service.threat.query.threat_publications(indicator)
        ti_pubs[indicator] = results
    # ti_pubs

    # correlate
    def check_tips(row, ti_pubs):
        for indicator, pub_list in ti_pubs.items():
            for col in [idx for idx in row.index if idx.startswith("logicals.")]:
                if pub_list and indicator in row[col]:
                    return pd.Series([True, pub_list])
        return pd.Series([False, []])

    df[["tips.found", "tips.publications"]] = df.apply(
        check_tips, ti_pubs=ti_pubs, axis=1
    )

    return df



import importlib.resources as pkg_resources
import inspect
import logging
import sys
import traceback
from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import List, Optional

import nbclient
import papermill
import typer
import yaml
from papermill.iorw import NoDatesSafeLoader, read_yaml_file
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResult
from taegis_magic.core.notebook import generate_report
from taegis_magic.core.service import get_service
from typing_extensions import Annotated

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Notebook Commands.")


@dataclass
class NotebookResult:
    action: str


class LOG_LEVEL(str, Enum):
    NOTSET = "NOTSET"
    DEBUG = "DEBUG"
    INFO = "INFO"
    WARNING = "WARNING"
    ERROR = "ERROR"
    CRITICAL = "CRITICAL"


@app.command()
@tracing
def execute(
    input_notebook: Annotated[
        Path,
        typer.Argument(
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
            help="Input notebook to parameterize and execute.",
        ),
    ],
    output_notebook: Annotated[
        Optional[Path],
        typer.Argument(
            exists=False,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
            help="Output notebook with parameters and results.",
        ),
    ] = None,
    parameter: Annotated[
        Optional[List[str]],
        typer.Option(
            "--parameter",
            "-p",
            help="Parameters to pass to the parameters cell. Use the format PARAMETER_NAME=VALUE.",
        ),
    ] = None,
    parameters_file: Annotated[
        Optional[List[Path]],
        typer.Option(
            "--parameters-file",
            "-f",
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
            help="YAML files to read parameters from. Can be used multiple times.",
        ),
    ] = None,
    parameters_yaml: Annotated[
        Optional[List[str]],
        typer.Option(
            "--parameters-yaml",
            "-y",
            help="YAML strings to read parameters from. Can be used multiple times.",
        ),
    ] = None,
    inject_input_path: Annotated[
        bool,
        typer.Option(
            "--inject-input-path",
            help="Insert the path of the input notebook as PAPERMILL_INPUT_PATH as a notebook parameter.",
        ),
    ] = False,
    inject_output_path: Annotated[
        bool,
        typer.Option(
            "--inject-output-path",
            help="Insert the path of the output notebook as PAPERMILL_OUTPUT_PATH as a notebook parameter.",
        ),
    ] = False,
    inject_paths: Annotated[
        bool,
        typer.Option(
            "--inject-paths",
            help="Insert the paths of input/output notebooks as PAPERMILL_INPUT_PATH/PAPERMILL_OUTPUT_PATH"
            " as notebook parameters.",
        ),
    ] = False,
    engine: Annotated[
        Optional[str],
        typer.Option(
            "--engine",
            help="The execution engine name to use in evaluating the notebook.",
        ),
    ] = None,
    request_save_on_cell_execute: Annotated[
        bool,
        typer.Option(
            "--request-save-on-cell-execute",
            help="Request save notebook after each cell execution",
        ),
    ] = True,
    autosave_cell_every: Annotated[
        int,
        typer.Option(
            "--autosave-cell-every",
            help="How often in seconds to autosave the notebook during long cell executions (0 to disable)",
        ),
    ] = 30,
    prepare_only: Annotated[
        bool,
        typer.Option(
            "--prepare-only/--prepare-execute",
            help="Flag for outputting the notebook without execution, but with parameters applied.",
        ),
    ] = False,
    kernel: Annotated[
        Optional[str],
        typer.Option(
            "--kernel",
            "-k",
            help="Name of kernel to run. Ignores kernel name in the notebook document metadata.",
        ),
    ] = None,
    language: Annotated[
        Optional[str],
        typer.Option(
            "--language",
            "-l",
            help="Language for notebook execution. Ignores language in the notebook document metadata.",
        ),
    ] = None,
    cwd: Annotated[
        Optional[str],
        typer.Option(
            "--cwd",
            help="Working directory to run notebook in.",
        ),
    ] = None,
    progress_bar: Annotated[
        bool,
        typer.Option(
            "--progress-bar/--no-progress-bar",
            help="Flag for turning on the progress bar.",
        ),
    ] = True,
    log_output: Annotated[
        bool,
        typer.Option(
            "--log-output/--no-log-output",
            help="Flag for writing notebook output to the configured logger.",
        ),
    ] = False,
    log_level: Annotated[
        LOG_LEVEL, typer.Option("--log-level", help="Log level to use.")
    ] = LOG_LEVEL.INFO,
    stdout_file: Annotated[
        Optional[typer.FileTextWrite],
        typer.Option(
            mode="w",
            encoding="utf-8",
            help="File to write notebook stdout output to.",
        ),
    ] = None,
    stderr_file: Annotated[
        Optional[typer.FileTextWrite],
        typer.Option(
            mode="w",
            encoding="utf-8",
            help="File to write notebook stderr output to.",
        ),
    ] = None,
    start_timeout: Annotated[
        int,
        typer.Option(
            "--start-timeout",
            help="Time in seconds to wait for the kernel to start.",
        ),
    ] = 60,
    execution_timeout: Annotated[
        int,
        typer.Option(
            "--execution-timeout",
            help="Time in seconds to wait for each cell before failing execution (default: forever)",
        ),
    ] = None,
    report_mode: Annotated[
        bool,
        typer.Option("--report-mode/--no-report-mode", help="Flag for hiding input."),
    ] = False,
    title: Annotated[Optional[str], typer.Option(help="Investigation Title.")] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Taegis Tenant ID.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Taegis Region.")] = None,
):
    """Execute a Jupyter notebook."""
    if not prepare_only:
        service = get_service(tenant_id=tenant, environment=region)
        service.access_token

    input_notebook = input_notebook or "-"

    if output_notebook is None:
        stem = input_notebook.stem
        if region:
            stem = f"{stem}_{region}"
        if tenant:
            stem = f"{stem}_{tenant}"

        stem = f"{stem}{input_notebook.suffix}"

        if stem != input_notebook.stem:
            output_notebook = input_notebook.with_name(stem)

    output_notebook = output_notebook or "-"

    if output_notebook == "-":
        # Save notebook to stdout just once
        request_save_on_cell_execute = False

        # Reduce default log level if we pipe to stdout
        if log_level == LOG_LEVEL.INFO:
            log_level = LOG_LEVEL.ERROR

    logging.basicConfig(level=log_level.value, format="%(message)s")

    parameters = {}
    bad_parameters = []
    if inject_input_path or inject_paths:
        parameters["PAPERMILL_INPUT_PATH"] = str(input_notebook)
    if inject_output_path or inject_paths:
        parameters["PAPERMILL_OUTPUT_PATH"] = str(output_notebook)
    for param in parameter or []:
        try:
            name, value = param.split("=", maxsplit=1)
        except ValueError:
            bad_parameters.append(param)
            continue

        parameters[name] = value
    for files in parameters_file or []:
        parameters.update(read_yaml_file(str(files)) or {})
    for params in parameters_yaml or []:
        parameters.update(yaml.load(params, Loader=NoDatesSafeLoader) or {})

    if title:
        parameters["INVESTIGATION_TITLE"] = title
    if tenant:
        parameters["TENANT_ID"] = tenant
    if region:
        parameters["REGION"] = region

    if bad_parameters:
        print(
            f"Please ensure the following parameters use the PARAMETER_NAME=VALUE format: {bad_parameters}"
        )
        raise typer.Exit()

    if "TAEGIS_MAGIC_NOTEBOOK_FILENAME" not in parameters:
        if output_notebook and output_notebook != "-":
            parameters["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = str(
                output_notebook.resolve()
            )
        else:
            parameters["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = str(input_notebook.resolve())

    try:
        papermill.execute_notebook(
            input_path=input_notebook,
            output_path=output_notebook,
            parameters=parameters,
            engine_name=engine,
            request_save_on_cell_execute=request_save_on_cell_execute,
            autosave_cell_every=autosave_cell_every,
            prepare_only=prepare_only,
            kernel_name=kernel,
            language=language,
            progress_bar=progress_bar,
            log_output=log_output,
            stdout_file=stdout_file,
            stderr_file=stderr_file,
            start_timeout=start_timeout,
            report_mode=report_mode,
            cwd=cwd,
            execution_timeout=execution_timeout,
        )
    except nbclient.exceptions.DeadKernelError:
        # Exiting with a special exit code for dead kernels
        traceback.print_exc()
        sys.exit(138)

    """Execute the current notebook."""
    return TaegisResult(
        raw_results=NotebookResult(action="execute"),
        service="notebook",
        tenant_id=None,
        region=None,
        arguments=inspect.currentframe().f_locals,
    )


@app.command(name="generate-report")
@tracing
def generate(input_notebook: Annotated[str, typer.Argument(...)]):
    """Generate a report from the current notebook."""
    report = generate_report(filename=input_notebook)

    return TaegisResult(
        raw_results=NotebookResult(action="generate_report"),
        service="notebook",
        tenant_id=None,
        region=None,
        arguments=inspect.currentframe().f_locals,
    )


@app.command()
@tracing
def create(
    output_notebook: Annotated[
        Path,
        typer.Argument(
            exists=False,
            file_okay=True,
            dir_okay=False,
            writable=True,
            readable=True,
            resolve_path=True,
        ),
    ],
):
    """Create a Jupyter Notebook to run against Taegis."""
    template = pkg_resources.read_text("taegis_magic.templates", "template.ipynb")

    output_notebook.write_text(template)

    return TaegisResult(
        raw_results=NotebookResult(action="create"),
        service="notebook",
        tenant_id=None,
        region=None,
        arguments=inspect.currentframe().f_locals,
    )



"""Pandas functions for Rules DataFrames."""

import logging

import pandas as pd

logger = logging.getLogger(__name__)


def inflate_filters(df: pd.DataFrame) -> pd.DataFrame:
    """Takes a DataFrame containing Taegis Event Filter
    Rules and concats an exploded JSON blob containing
    the Event Filter criteria.

    Parameters
    ----------
    df : pd.DataFrame
        Rules DataFrame

    Returns
    -------
    pd.DataFrame
        Inflated Rules DataFrame

    Raises
    ------
    ValueError
        DataFrame must have a 'filters' column.

    Example
    -------
    Example::

        %taegis rules suppression --event-type cloudaudit --assign df

        inflated_df = df.pipe(inflate_filters)
    """
    if "filters.id" in df.columns:
        return df

    if "filters" not in df.columns:
        raise ValueError("DataFrame must have a 'filters' column.")

    df = df.explode("filters").reset_index(drop=True)
    return pd.concat(
        [
            df,
            pd.json_normalize(
                df["filters"].apply(lambda x: {} if pd.isnull(x) else x).to_list(),
                max_level=3,
            ).add_prefix("filters."),
        ],
        axis=1,
    )



"""Taegis Magic search commands."""

import logging
from typing import Optional

import typer
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResults
from taegis_magic.core.service import get_service
from taegis_sdk_python.services.nl_search.types import NLSearchInputs
from typing_extensions import Annotated

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Search Commands.")


@app.command()
@tracing
def generate(
    cell: Annotated[
        str, typer.Option(help="Natural language query to convert to Taegis QL")
    ],
    limit: Annotated[int, typer.Option(help="Limit number of results")] = 5,
    region: Annotated[Optional[str], typer.Option(help="Taegis Region")] = None,
):
    """Generate Taegis QL from natural language query."""
    service = get_service(environment=region)

    results = service.nl_search.query.nl_search(
        in_=NLSearchInputs(
            query=cell,
            limit=limit,
        )
    )

    return TaegisResults(
        raw_results=results,
        service="search",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments={
            "cell": cell,
            "limit": limit,
            "region": region,
        },
    )



import logging
import os
import re
import sys
from concurrent.futures import ProcessPoolExecutor, as_completed
from dataclasses import dataclass
from importlib.metadata import version as module_version
from pathlib import Path
from time import sleep
from typing import List, Optional, Union

import jinja2
import nbconvert
import papermill
from ipylab import JupyterFrontEnd
from IPython import get_ipython
from IPython.display import HTML, Javascript, display
from nbconvert.preprocessors.tagremove import TagRemovePreprocessor
from packaging.version import Version
from packaging.version import parse as parse_version
from traitlets.config import Config

log = logging.getLogger(__name__)

VERSION_6 = parse_version("6")
VERSION_7 = parse_version("7")
VERSION_8 = parse_version("8")


def pool_initializer():
    """Default ProcessPoolExecutor initializer.

    https://github.com/ipython/ipython/issues/11049
    """
    
    sys.stdout.write(' ')
    sys.stdout.flush()

@dataclass
class NotebookContext:
    """Metadata for a Jupyter notebook."""

    tenant: str
    region: str
    parameters: Optional[dict] = None


@dataclass
class NotebookScope:
    """Scope for a Jupyter notebook execution."""

    tenant: str
    region: str
    notebook_title: Optional[str] = None
    notebook_path: Optional[Union[str, Path]] = None


def get_notebook_version() -> Version:
    return parse_version(module_version("notebook"))


def find_notebook_name() -> Optional[str]:
    """Find the name of the current notebook."""
    notebook_name = None

    if not notebook_name:
        ip = get_ipython()
        if notebook_name := ip.user_ns.get("__vsc_ipynb_file__"):
            notebook_name = Path(notebook_name).name
            log.debug(f"Notebook found via vscode: {notebook_name}")
        else:
            log.debug("Could not find notebook name using __vsc_ipynb_file__")

    if not notebook_name:
        if notebook_name := os.environ.get("JPY_SESSION_NAME"):
            notebook_name = Path(notebook_name).name
            log.debug(f"Notebook found via JPY_SESSION_NAME: {notebook_name}")
        else:
            log.debug("Could not find notebook name using JPY_SESSION_NAME")

    if not notebook_name:
        try:
            # not best practice, but there are some load time issues
            # with ipynbname that we need to work through
            # if we don't need to load it, we shouldn't
            import ipynbname

            notebook_name = ipynbname.path().name
            log.debug(f"Notebook found via ipynbname: {notebook_name}")
        except Exception as e:
            log.debug(f"Error finding notebook name using ipynbname: {e}")

    return notebook_name


def save_notebook(delay: int = 0):
    """Save the current notebook."""
    ip = get_ipython()
    if "__vsc_ipynb_file__" in ip.user_ns:
        log.error(
            "save_notebook does not work in VS Code notebooks, please save manually before proceeding."
        )
        return

    notebook_version = get_notebook_version()
    log.debug(f"Notebook version: {notebook_version}")

    if notebook_version >= VERSION_6 and notebook_version < VERSION_7:
        display('<-- #region tags=["remove_cell"] -->')

        try:
            script = """
            this.nextElementSibling.focus();
            this.dispatchEvent(new KeyboardEvent('keydown', {key:'s', keyCode: 83, ctrlKey: true}));
            """
            display(
                HTML(
                    (
                        '<img src onerror="{}" style="display:none">'
                        '<input style="width:0;height:0;border:0">'
                    ).format(script)
                )
            )
        except Exception:
            pass

        try:
            display(Javascript("IPython.notebook.save_checkpoint();"))
        except Exception:
            pass

        display("<-- #endregion -->")

    elif notebook_version >= VERSION_7 and notebook_version < VERSION_8:
        try:
            app = JupyterFrontEnd()
            app.commands.execute("docmanager:save")
        except Exception:
            pass

    else:
        log.error(
            "Cannot save notebook, unsupported notebook version.  Please save manually before proceeding."
        )
        return

    sleep(delay)


def remove_region_tags(body: str) -> str:
    """Remove region tags from a report that may not be related to cells."""
    region_pattern = (
        r'\'<--\s*#region\s*tags=\["remove_cell"\]\s*-->(.*?)<--\s*#endregion\s*-->\''
    )
    body = re.sub(region_pattern, "", body, flags=re.DOTALL)
    return body


def generate_report(filename: Union[str, Path]) -> Path:
    """Takes path to Jupyter notebook and uses nbconvert
    to export notebook as markdown.

    Uses nbconvert preprocessors to allow users to
    manually tag notebook cells for inclusion/exclusion
    in the output.

    Also uses jinja templates to override notebook
    output. By default, all code cells, cell input,
    raw cells, and streaming output is excluded.
    These templates live in `common.templates`.

    Parameters
    ----------
    filename : str | Path
        Notebook file name

    Raises
    ------
    FileNotFoundError
        If the file does not exist
    """

    # Configure our tag removal. I believe we can add/change
    # the cell tags that will trigger include/exclude behavior
    # in the tuples below.
    # save_notebook()

    log.debug("Configuring nbconvert preprocessor...")

    if isinstance(filename, str):
        filename = Path(filename)

    if not filename.exists():
        raise FileNotFoundError(f"File {filename} does not exist")

    c = Config()
    c["TagRemovePreprocessor"].remove_cell_tags = ("remove_cell",)
    c["TagRemovePreprocessor"].remove_all_outputs_tags = ("remove_output",)
    c["TagRemovePreprocessor"].remove_input_tags = ("remove_input",)

    c.MarkdownExporter.preprocessors = [TagRemovePreprocessor]
    exporter = nbconvert.MarkdownExporter(config=c)
    template_file = "jupyter_extended_markdown_template.jinja"

    # The nbconvert exporter has its own jinja environment.
    # We need to point that environment to our templates
    # directory to allow them to be loaded. Otherwise defaults
    # to the local directory and off Jupyter path
    exporter.environment.loader.loaders.append(
        jinja2.PackageLoader("taegis_magic", "templates")
    )

    # log.debug(exporter.environment.loader.list_templates())

    # Not sure how this interops with `report_mode` in `papermill`
    exporter.exclude_input_prompt = True
    exporter.exclude_output_prompt = True
    exporter.exclude_input = True
    exporter.exclude_raw = True
    exporter.template_file = template_file

    output_file = filename.with_suffix(".report.md")
    body, _ = exporter.from_filename(filename)

    # clear output from --cache
    body = remove_region_tags(body)

    log.info(f"Writing markdown output to {str(output_file.resolve())}")
    Path(output_file).write_text(body, encoding="utf-8", errors="replace")

    return output_file


def execute_notebook_pool(
    notebook_context: List[NotebookContext],
    notebook_path: Union[str, Path],
    notebook_title: str,
    process_pool_kwargs: Optional[dict] = None,
    papermill_kwargs: Optional[dict] = None,
):
    """Execute a notebook in parallel using papermill and ProcessPoolExecutor.

    Parameters
    ----------
    notebook_context : List[NotebookContext]
        List of notebook contexts to execute.
    notebook_path : Union[str, Path]
        File path to notebook to execute.
    notebook_title : str
        Case Title for the notebook.
    process_pool_kwargs : Optional[dict], optional
        ProcessPoolExecutor keyword arguments, by default None
    papermill_kwargs : Optional[dict], optional
        Papermill keyword arguments, by default None

    Raises
    ------
    FileNotFoundError
    """
    if process_pool_kwargs is None:
        process_pool_kwargs = {}
    if papermill_kwargs is None:
        papermill_kwargs = {}

    if process_pool_kwargs.get("initializer") is None:
        process_pool_kwargs["initializer"] = pool_initializer

    if isinstance(notebook_path, str):
        notebook_path = Path(notebook_path)
    if not notebook_path.exists():
        raise FileNotFoundError(f"Notebook {notebook_path} does not exist")

    executed_procedures = {}
    with ProcessPoolExecutor(**process_pool_kwargs) as executor:
        for context in notebook_context:
            if papermill_kwargs.get("output_notebook") is None:
                stem = notebook_path.stem
                if context.region:
                    stem = f"{stem}_{context.region}"
                if context.tenant:
                    stem = f"{stem}_{context.tenant}"

                stem = f"{stem}{notebook_path.suffix}"

                if stem != notebook_path.stem:
                    output_notebook = notebook_path.with_name(stem)

            parameters = context.parameters or {}

            if (
                papermill_kwargs.get("inject_input_path")
                or papermill_kwargs.get("inject_paths")
            ):
                parameters["PAPERMILL_INPUT_PATH"] = str(notebook_path)
            if (
                papermill_kwargs.get("inject_output_path")
                or papermill_kwargs.get("inject_paths")
            ):
                parameters["PAPERMILL_OUTPUT_PATH"] = str(output_notebook)

            if notebook_title:
                parameters["INVESTIGATION_TITLE"] = notebook_title
            if context.tenant:
                parameters["TENANT_ID"] = context.tenant
            if context.region:
                parameters["REGION"] = context.region

            if "TAEGIS_MAGIC_NOTEBOOK_FILENAME" not in parameters:
                if output_notebook and output_notebook != "-":
                    parameters["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = str(
                        output_notebook.resolve()
                    )
                else:
                    parameters["TAEGIS_MAGIC_NOTEBOOK_FILENAME"] = str(
                        notebook_path.resolve()
                    )

            executed_procedures[
                executor.submit(
                    papermill.execute_notebook,
                    input_path=notebook_path,
                    output_path=output_notebook,
                    parameters=parameters,
                    **papermill_kwargs,
                )
            ] = NotebookScope(
                tenant=context.tenant,
                region=context.region,
                notebook_title=notebook_title,
                notebook_path=notebook_path,
            )

    # Wait for the notebooks to finish running
    # and log exceptions as they occur.
    for i, executed_procedure in enumerate(as_completed(executed_procedures)):
        procedure_scope = executed_procedures[executed_procedure]
        log.debug(
            f"{i+1}/{len(executed_procedures.keys())} {procedure_scope} Finished!"
        )
        exc = executed_procedure.exception()
        if exc:
            log.error(procedure_scope, exc)



"""Taegis Magic Command callback utilities."""

from pathlib import Path

import typer


def verify_file(value: str):
    """Verify a file on disk."""
    fp = Path(value)
    overwrite = None
    options = ["y", "n"]

    if fp.exists():
        while not overwrite in options:
            overwrite = input(f"{fp} exists.  Overwrite [{'/'.join(options)}]? ")

        if overwrite == "n":
            raise typer.Exit()

    return fp



"""Taegis Magic CLI definitions."""

import json
import logging
import sys

import typer

from gql.transport.exceptions import TransportQueryError
from taegis_magic.commands import (
    alerts,
    audits,
    authentication,
    clients,
    configure,
    events,
    investigations,
    notebook,
    preferences,
    rules,
    search,
    sharelinks,
    subjects,
    tenant_profiles,
    tenants,
    threat,
    users,
    process_trees,
)
from taegis_magic.core.normalizer import TaegisResult
from dataclasses_json import dataclass_json
from dataclasses import dataclass

from taegis_magic.core.log import TRACE_LOG_LEVEL, get_module_logger

log = logging.getLogger(__name__)


@dataclass_json
@dataclass
class MagicVersion:
    python: str
    taegis_sdk_python: str
    taegis_magic: str


CONTEXT_SETTINGS = dict(help_option_names=["--help", "-h"])

app = typer.Typer(context_settings=CONTEXT_SETTINGS)
app.add_typer(alerts.app, name="alerts")
app.add_typer(authentication.app, name="auth")
app.add_typer(audits.app, name="audits")
app.add_typer(clients.app, name="clients")
app.add_typer(configure.app, name="configure")
app.add_typer(events.app, name="events")
app.add_typer(investigations.app, name="investigations")
app.add_typer(notebook.app, name="notebook")
app.add_typer(preferences.app, name="preferences")
app.add_typer(rules.app, name="rules")
app.add_typer(search.app, name="search")
app.add_typer(sharelinks.app, name="sharelinks")
app.add_typer(subjects.app, name="subjects")
app.add_typer(tenant_profiles.app, name="tenant-profiles")
app.add_typer(tenants.app, name="tenants")
app.add_typer(threat.app, name="threat")
app.add_typer(users.app, name="users")
app.add_typer(process_trees.app, name="process-trees")

CONFIG = configure.set_defaults()


@app.callback()
def main(
    warning: bool = CONFIG[configure.LOGGING_SECTION].getboolean(
        "warning", fallback=True
    ),
    verbose: bool = CONFIG[configure.LOGGING_SECTION].getboolean(
        "verbose", fallback=False
    ),
    debug: bool = CONFIG[configure.LOGGING_SECTION].getboolean("debug", fallback=False),
    trace: bool = CONFIG[configure.LOGGING_SECTION].getboolean("trace", fallback=False),
    sdk_warning: bool = CONFIG[configure.LOGGING_SECTION].getboolean(
        "sdk_warning", fallback=True
    ),
    sdk_verbose: bool = CONFIG[configure.LOGGING_SECTION].getboolean(
        "sdk_verbose", fallback=False
    ),
    sdk_debug: bool = CONFIG[configure.LOGGING_SECTION].getboolean(
        "sdk_debug", fallback=False
    ),
):
    """Taegis Magic help menu."""
    logger = get_module_logger()
    if trace:
        logger.setLevel(TRACE_LOG_LEVEL)
    elif debug:
        logger.setLevel(logging.DEBUG)
    elif verbose:
        logger.setLevel(logging.INFO)
    elif warning:
        logger.setLevel(logging.WARNING)
    else:
        logger.setLevel(logging.ERROR)

    sdk_logger = logging.getLogger("taegis_sdk_python")

    if sdk_debug:
        sdk_logger.setLevel(logging.DEBUG)
    elif sdk_verbose:
        sdk_logger.setLevel(logging.INFO)
    elif sdk_warning:
        sdk_logger.setLevel(logging.WARNING)
    else:
        sdk_logger.setLevel(logging.ERROR)


@app.command()
def version():
    """Taegis Magic version information."""
    import sys
    from taegis_sdk_python._version import __version__ as sdk_version
    from taegis_magic._version import __version__ as magic_version

    return TaegisResult(
        raw_results=MagicVersion(
            python=sys.version,
            taegis_sdk_python=sdk_version,
            taegis_magic=magic_version,
        ),
        tenant_id=None,
        region=None,
        service="version",
    )


def cli():
    """
    Run app as a cli command.
    """
    try:
        result = app(standalone_mode=False)
    except TransportQueryError as exc:
        print(json.dumps(exc.errors))
        sys.exit(1)
    except SystemExit:
        sys.exit(0)

    if isinstance(result, int):
        sys.exit(result)

    print(json.dumps(result.results))


if __name__ == "__main__":
    app()



"""Magics Configuration tool."""

import logging
from dataclasses import dataclass
from enum import Enum
from typing import List

import typer
from dataclasses_json import dataclass_json
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import TaegisResultsNormalizer
from typing_extensions import Annotated

from taegis_sdk_python.config import get_config, write_config, write_to_config

log = logging.getLogger(__name__)

app = typer.Typer(help="Taegis Magic Configuration Commands.")

auth = typer.Typer(help="Configure Authentication options.")
regions = typer.Typer(help="Configure Addtional Regions Commands.")
queries = typer.Typer(help="Configure Default Query Commands.")
configure_logging = typer.Typer(help="Configure Magic Logging Commands.")

app.add_typer(auth, name="auth")
app.add_typer(regions, name="regions")
app.add_typer(queries, name="queries")
app.add_typer(configure_logging, name="logging")

AUTH_SECTION = "magics.auth"
REGIONS_SECTION = "magics.regions"
QUERIES_SECTION = "magics.queries"
LOGGING_SECTION = "magics.logging"


class UseUniversalAuthOptions(str, Enum):
    """Configure default logging options."""

    true = "true"
    false = "false"


class LoggingOptions(str, Enum):
    """Configure Taegis Magic logging."""

    trace = "trace"
    debug = "debug"
    verbose = "verbose"
    warning = "warning"
    sdk_debug = "sdk_debug"
    sdk_verbose = "sdk_verbose"
    sdk_warning = "sdk_warning"


class ConfigureLogging(str, Enum):
    """Configure default logging options."""

    true = "true"
    false = "false"


class QueriesTrack(str, Enum):
    """Automatically track alert/event search queries."""

    yes = "yes"
    no = "no"


@dataclass_json
@dataclass
class ConfigurationNormalizer(TaegisResultsNormalizer):
    """Configuration Normalizer."""


@tracing
def set_defaults():  # pragma: no cover
    """Set default configuration values for Taegic Magics."""
    config = get_config()

    ###
    # Add Sections
    ###
    if not config.has_section(AUTH_SECTION):
        config.add_section(AUTH_SECTION)

    if not config.has_section(REGIONS_SECTION):
        config.add_section(REGIONS_SECTION)

    if not config.has_section(QUERIES_SECTION):
        config.add_section(QUERIES_SECTION)

    if not config.has_section(LOGGING_SECTION):
        config.add_section(LOGGING_SECTION)

    ###
    # Auth Defaults
    ###
    if not config.has_option(AUTH_SECTION, "use_universal_auth"):
        config[AUTH_SECTION]["use_universal_auth"] = "false"

    ###
    # Queries Defaults
    ###
    if not config.has_option(QUERIES_SECTION, "track"):
        config[QUERIES_SECTION]["track"] = "no"

    ###
    # Logging Defaults
    ###
    if not config.has_option(LOGGING_SECTION, LoggingOptions.warning):
        config[LOGGING_SECTION][LoggingOptions.warning.value] = "true"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.verbose):
        config[LOGGING_SECTION][LoggingOptions.verbose.value] = "false"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.debug):
        config[LOGGING_SECTION][LoggingOptions.debug.value] = "false"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.trace):
        config[LOGGING_SECTION][LoggingOptions.trace.value] = "false"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.sdk_warning):
        config[LOGGING_SECTION][LoggingOptions.sdk_warning.value] = "true"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.sdk_verbose):
        config[LOGGING_SECTION][LoggingOptions.sdk_verbose.value] = "false"
    if not config.has_option(LOGGING_SECTION, LoggingOptions.sdk_debug):
        config[LOGGING_SECTION][LoggingOptions.sdk_debug.value] = "false"

    write_config(config)

    return config


@auth.command("list")
@tracing
def auth_list():
    """Configure use of Universal Authentication."""
    config = get_config()

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[
            {
                name: value,
            }
            for name, value in config[AUTH_SECTION].items()
        ],
    )
    return results


@auth.command("use-universal-auth")
@tracing
def use_universal_auth(
    value: Annotated[
        UseUniversalAuthOptions,
        typer.Argument(help="Use Universal Authentication for Taegis Magic."),
    ] = UseUniversalAuthOptions.false
):
    """Configure use of Universal Authentication."""
    config = get_config()

    config[AUTH_SECTION]["use_universal_authentication"] = value.value

    write_config(config)

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(use_universal_authentication=value.value)],
    )
    return results


@regions.command(name="add")
@tracing
def add_regions(
    name: str,
    url: str,
):
    """Configure addition Taegis regions for magics."""
    write_to_config(
        REGIONS_SECTION,
        name,
        url,
    )

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(name=name, url=url)],
    )
    return results


@regions.command(name="remove")
@tracing
def remove_regions(name: str):
    """Remove additional Taegis regions from configuration."""
    config = get_config()

    config.remove_option(REGIONS_SECTION, name)

    write_config(config)

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(name=name)],
    )
    return results


@regions.command(name="list")
@tracing
def list_regions():
    """List additional regions."""
    config = get_config()

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[
            dict(
                name=name,
                url=url,
            )
            for name, url in config[REGIONS_SECTION].items()
        ],
    )
    return results


@queries.command(name="track")
@tracing
def queries_track(status: QueriesTrack = QueriesTrack.no):
    """Configure Alert/Event query tracking by default."""
    write_to_config(
        QUERIES_SECTION,
        "track",
        status.value,
    )

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(status=status.value)],
    )
    return results


@queries.command(name="list")
@tracing
def queries_list():
    """List queries configurations."""
    config = get_config()

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[
            dict(
                name=name,
                value=value,
            )
            for name, value in config[QUERIES_SECTION].items()
        ],
    )
    return results


@queries.command(name="callername")
@tracing
def callername(
    name: Annotated[
        str,
        typer.Argument(
            help="Caller Name to use for Alert/Event queries; used for searching/filtering results from the Queries API."
        ),
    ]
):
    """Configure Alert/Event query callername by default."""
    write_to_config(
        QUERIES_SECTION,
        "callername",
        name,
    )

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(callername=name)],
    )
    return results


@configure_logging.command(name="defaults")
@tracing
def logging_defaults(
    option: LoggingOptions, status: Annotated[ConfigureLogging, typer.Option()]
):
    """Configure logging defaults."""
    write_to_config(
        LOGGING_SECTION,
        option.value,
        status.value,
    )

    results = ConfigurationNormalizer(
        service="configure",
        tenant_id="None",
        region="None",
        raw_results=[dict(option=option.value, status=status.value)],
    )
    return results



"""Taegis Magic investigations commands."""

import inspect
import logging
import mimetypes
import re
import time
from dataclasses import asdict, dataclass, field
from enum import Enum
from pathlib import Path
from textwrap import dedent
from typing import Any, Dict, List, Optional, Union
from itertools import product

import requests
import typer
from dataclasses_json import config, dataclass_json
from taegis_magic.commands.utils.investigations import (
    InvestigationEvidenceNormalizer,
    InvestigationEvidenceType,
    InvestigationEvidenceChanges,
    clear_search_queries,
    delete_search_query,
    find_database,
    find_dataframe,
    get_investigation_evidence,
    insert_search_query,
    list_search_queries,
    lookup_assignee_id,
    read_database,
    stage_investigation_evidence,
    unstage_investigation_evidence,
)
from taegis_magic.core.callbacks import verify_file
from taegis_magic.core.log import tracing
from taegis_magic.core.normalizer import (
    DataFrameNormalizer,
    TaegisResult,
    TaegisResults,
    TaegisResultsNormalizer,
)
from taegis_magic.core.service import get_service
from taegis_magic.core.utils import remove_output_node
from typing_extensions import Annotated

from taegis_sdk_python import (
    GraphQLNoRowsInResultSetError,
    GraphQLService,
    build_output_string,
    prepare_input,
)
from taegis_sdk_python.services.investigations2.types import (
    AddCommentToInvestigationInput,
    AddEvidenceToInvestigationInput,
    RemoveEvidenceFromInvestigationInput,
    ArchiveInvestigationInput,
    CreateInvestigationInput,
    DeleteInvestigationFileInput,
    InitInvestigationFileUploadInput,
    InvestigationFilesV2Arguments,
    InvestigationFileV2Arguments,
    InvestigationStatus,
    InvestigationsV2,
    InvestigationsV2Arguments,
    InvestigationType,
    InvestigationV2,
    InvestigationV2Arguments,
    UpdateInvestigationV2Input,
    CommentsV2Arguments,
    AddCommentToInvestigationInput,
    UpdateInvestigationCommentInput,
    DeleteInvestigationCommentInput,
)
from taegis_sdk_python.services.queries.types import QLQueriesInput
from taegis_sdk_python.services.sharelinks.types import ShareLinkCreateInput
from taegis_sdk_python.services.subjects.types import Subject as FederatedSubject

log = logging.getLogger(__name__)


app = typer.Typer(help="Taegis Investigation Commands.")
investigations_attachment = typer.Typer(help="Investigation File Attachment commands.")
investigations_comment = typer.Typer(help="Investigation Comment commands.")
investigations_evidence = typer.Typer(help="Investigation Evidence commands.")
investigations_search_queries = typer.Typer(help="Investigation Search Query commands.")

app.add_typer(
    investigations_attachment,
    name="attachment",
)
app.add_typer(
    investigations_comment,
    name="comment",
)
app.add_typer(
    investigations_evidence,
    name="evidence",
)
app.add_typer(
    investigations_search_queries,
    name="search-queries",
)


class InvestigationPriority(str, Enum):
    LOW = "LOW"
    MEDIUM = "MEDIUM"
    HIGH = "HIGH"
    CRITICAL = "CRITICAL"


INVESTIGATION_PRIORITY_MAP = {
    InvestigationPriority.LOW: 1,
    InvestigationPriority.MEDIUM: 2,
    InvestigationPriority.HIGH: 3,
    InvestigationPriority.CRITICAL: 4,
}


@dataclass_json
@dataclass
class InsertSearchQueryNormalizer:
    """Taegis Normalizer Query Normalizer (Duck Typed)."""

    query_identifier: str
    tenant_id: str
    query: str
    results_returned: int
    total_results: int


@dataclass_json
@dataclass
class InvestigationsSearchResultsNormalizer(TaegisResultsNormalizer):
    """Investigations Results Normalizer."""

    raw_results: Optional[List[InvestigationsV2]] = None
    _shareable_url: List[str] = field(default_factory=list)

    def __post_init__(self):
        self._shareable_url = [None for _ in range(self.results_returned)]

    @property
    def results(self) -> List[Dict[str, Any]]:
        """Query results from Investigations search.

        Returns
        -------
        List[Dict[str, Any]]
            List of results.
        """
        return (
            [
                asdict(investigation)
                for result in self.raw_results
                for investigation in result.investigations
            ]
            if self.raw_results
            else []
        )

    @property
    def status(self) -> str:
        """Status of query.

        Returns
        -------
        str
            Status message of query.
        """
        return "ERROR" if self.raw_results is None else "SUCCESS"

    @property
    def total_results(self) -> int:
        """Total number found by API.

        Returns
        -------
        int
            Total number found by API.
        """
        return -1 if self.raw_results is None else self.raw_results[0].total_count

    @property
    def results_returned(self) -> int:
        """Total number returned by API.

        Returns
        -------
        int
            Total number returned by API.
        """
        return -1 if self.raw_results is None else len(self.results)

    def get_shareable_url(self, index: int = 0) -> str:
        """Query Shareable Link.

        Returns
        -------
        str
            Returns a link created by the Preferences API.
        """
        if self.raw_results is None:
            return "No share link producible"

        investigation = self.results[index]

        if self._shareable_url[index]:
            return self._shareable_url[index]

        service = get_service(environment=self.region, tenant_id=self.tenant_id)

        result = service.sharelinks.mutation.create_share_link(
            ShareLinkCreateInput(
                link_ref=investigation.get("id"),
                link_type="investigationId",
                tenant_id=self.tenant_id,
            )
        )

        self._shareable_url[index] = (
            service.investigations.sync_url.replace("api.", "") + f"/share/{result.id}"
        )
        return self._shareable_url[index]


@dataclass_json
@dataclass
class InvestigationsCreatedResultsNormalizer(TaegisResultsNormalizer):
    """Investigations Results Normalizer."""

    raw_results: Union[InvestigationV2, CreateInvestigationInput] = field(
        default_factory=lambda: InvestigationV2()
    )
    dry_run: bool = False

    _shareable_url: Optional[str] = None

    @property
    def results(self):
        return [asdict(self.raw_results)]

    @property
    def status(self) -> str:
        """Status of query.

        Returns
        -------
        str
            Status message of query.
        """
        if self.dry_run:
            return "DRY_RUN"

        if self.raw_results:
            return "SUCCESS"

        return "ERROR"

    @property
    def total_results(self) -> int:
        """Total number found by API.

        Returns
        -------
        int
            Total number found by API.
        """
        return 0 if self.raw_results is None else 1

    @property
    def results_returned(self) -> int:
        """Total number returned by API.

        Returns
        -------
        int
            Total number returned by API.
        """
        return -1 if self.raw_results is None else len(self.results)

    def _repr_markdown_(self):
        if self.dry_run:
            return dedent(
                f"""
                Dry Run:

                ```json
                {self.raw_results.to_json()}
                ```
                """
            )
        else:
            return dedent(
                f"""
                | Investigation ID  | Short ID                | Title                | Type                | Share Link           |
                | ----------------- | ----------------------- | -------------------- | ------------------- | -------------------- |
                | {self.raw_results.id} | {self.raw_results.short_id} | {self.raw_results.title} | {self.raw_results.type} | {self.shareable_url} |
                """
            )

    @property
    def shareable_url(self) -> str:
        """Create a Shareable URL.

        Returns
        -------
        str
            Returns a link created by the Sharelinks API.
        """
        if self._shareable_url:
            return self._shareable_url

        if not isinstance(self.raw_results, InvestigationV2):
            return "Not Available"

        investigation = self.raw_results

        service = get_service(environment=self.region, tenant_id=self.tenant_id)

        result = service.sharelinks.mutation.create_share_link(
            ShareLinkCreateInput(
                link_ref=investigation.id,
                link_type="investigationId",
                tenant_id=self.tenant_id,
            )
        )

        self._shareable_url = (
            service.investigations.sync_url.replace("api.", "") + f"/share/{result.id}"
        )

        return self._shareable_url


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class TaegisMagicInvestigationV2(InvestigationV2):
    contributor_subjects: Optional[List[FederatedSubject]] = field(
        default=None, metadata=config(field_name="contributorSubjects")
    )
    assignee_subject: Optional[FederatedSubject] = field(
        default=None, metadata=config(field_name="assigneeSubject")
    )
    created_by_subject: Optional[FederatedSubject] = field(
        default=None, metadata=config(field_name="createdBySubject")
    )
    updated_by_subject: Optional[FederatedSubject] = field(
        default=None, metadata=config(field_name="updatedBySubject")
    )


@dataclass_json
@dataclass(order=True, eq=True, frozen=True)
class TaegisMagicInvestigationsV2(InvestigationsV2):
    investigations: List[TaegisMagicInvestigationV2] = field(
        default_factory=list, metadata=config(field_name="investigations")
    )


def federated_investigation_create(
    service: GraphQLService, input_: CreateInvestigationInput
) -> TaegisMagicInvestigationV2:
    """createInvestigationV2 creates new investigation with the provided arguments.."""
    endpoint = "createInvestigationV2"

    result = service.investigations2.execute_mutation(
        endpoint=endpoint,
        variables={
            "input": prepare_input(input_),
        },
        output=build_output_string(TaegisMagicInvestigationV2),
    )
    if result.get(endpoint) is not None:
        return TaegisMagicInvestigationV2.from_dict(result.get(endpoint))
    raise GraphQLNoRowsInResultSetError("for mutation createInvestigationV2")


def federated_investigations_search(
    service, arguments: InvestigationsV2Arguments
) -> TaegisMagicInvestigationsV2:
    """investigationsV2 returns a list of investigations matching the provided arguments."""
    endpoint = "investigationsV2"

    result = service.investigations2.execute_query(
        endpoint=endpoint,
        variables={
            "arguments": prepare_input(arguments),
        },
        output=build_output_string(TaegisMagicInvestigationsV2),
    )
    if result.get(endpoint) is not None:
        return TaegisMagicInvestigationsV2.from_dict(result.get(endpoint))
    raise GraphQLNoRowsInResultSetError("for query investigationsV2")


@investigations_evidence.command(name="stage")
@tracing
def evidence_stage(
    evidence_type: Annotated[
        Optional[InvestigationEvidenceType],
        typer.Argument(
            help="InvestigationType; will gather type from DataFrame if not provided."
        ),
    ],
    dataframe: Annotated[str, typer.Argument(help="Data Reference.")],
    database: Annotated[
        str, typer.Option(help="Database reference.  Provide a file path or :memory:")
    ] = ":memory:",
    investigation_id: Annotated[
        str, typer.Option(help="Taegis Investigation Identifier.")
    ] = "NEW",
):
    """
    Stage evidence prior to linking to an investigation.
    """
    df = find_dataframe(dataframe)
    db = find_database(database)

    if evidence_type == evidence_type.All:
        if "taegis_magic.evidence_type" not in df.columns:
            raise ValueError(
                "DataFrame must contain 'taegis_magic.evidence_type' column to stage all evidence types."
            )

        changes = InvestigationEvidenceChanges(
            action="stage", evidence_type="All", investigation_id=investigation_id
        )

        for evidence_type, investigation_id in product(
            df["taegis_magic.evidence_type"].unique(), df["investigation_id"].unique()
        ):
            stage_df = df[
                (
                    (df["taegis_magic.evidence_type"] == evidence_type)
                    & (df["investigation_id"] == investigation_id)
                )
            ]

            type_changes = stage_investigation_evidence(
                stage_df,
                db,
                evidence_type,
                investigation_id,
            )

            changes.before += type_changes.before
            changes.after += type_changes.after
            changes.difference += type_changes.difference
            log.debug(f"Staged evidence type ({evidence_type}): {changes.to_json()}")
    else:
        changes = stage_investigation_evidence(df, db, evidence_type, investigation_id)

    return InvestigationEvidenceNormalizer(
        raw_results=changes,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )


@investigations_evidence.command(name="unstage")
@tracing
def evidence_unstage(
    evidence_type: Annotated[
        InvestigationEvidenceType, typer.Argument(help="Investigation Evidence Type.")
    ],
    dataframe: Annotated[str, typer.Argument(help="Data Reference.")],
    database: Annotated[
        str,
        typer.Option(help="Local database file.  Use :memory: for in-memory database."),
    ] = ":memory:",
    investigation_id: Annotated[
        str, typer.Option(help="Investigation Identifier.")
    ] = "NEW",
):
    """
    Remove staged evidence prior to linking to an investigation.
    """

    df = find_dataframe(dataframe)
    db = find_database(database)

    df["investigation_id"].fillna(investigation_id, inplace=True)

    if evidence_type == evidence_type.All:
        if "taegis_magic.evidence_type" not in df.columns:
            raise ValueError(
                "DataFrame must contain 'taegis_magic.evidence_type' column to unstage all evidence types."
            )

        changes = InvestigationEvidenceChanges(
            action="unstage", evidence_type="All", investigation_id=investigation_id
        )

        for evidence_type, investigation_id in product(
            df["taegis_magic.evidence_type"].unique(), df["investigation_id"].unique()
        ):
            stage_df = df[
                (
                    (df["taegis_magic.evidence_type"] == evidence_type)
                    & (df["investigation_id"] == investigation_id)
                )
            ]

            type_changes = unstage_investigation_evidence(
                stage_df,
                db,
                evidence_type,
                investigation_id,
            )

            changes.before += type_changes.before
            changes.after += type_changes.after
            changes.difference += type_changes.difference
            log.debug(f"Staged evidence type ({evidence_type}): {changes.to_json()}")
    else:
        changes = unstage_investigation_evidence(
            df, db, evidence_type, investigation_id
        )

    return InvestigationEvidenceNormalizer(
        raw_results=changes,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )


@investigations_evidence.command(name="clear")
@tracing
def evidence_clear(
    database: Annotated[
        str,
        typer.Option(help="Local database file.  Use :memory: for in-memory database."),
    ] = ":memory:",
    investigation_id: Annotated[
        str, typer.Option(help="Investigation Identifier.")
    ] = "NEW",
    tenant: Annotated[
        Optional[str], typer.Option(help="Taegis Tenant Identifier.")
    ] = None,
):
    """
    Clear all currently staged evidence.
    """
    db = find_database(database)

    for evidence_type in InvestigationEvidenceType:
        df = read_database(
            db,
            evidence_type=evidence_type,
            tenant_id=tenant,
            investigation_id=investigation_id,
        )
        log.debug(f"Found evidence type ({evidence_type}): {df.to_markdown()}")
        changes = unstage_investigation_evidence(
            df, db, evidence_type, investigation_id
        )
        log.debug(f"Unstaged evidence: {changes.to_json()}")

    df = read_database(
        db,
        evidence_type=None,
        tenant_id=tenant,
        investigation_id=investigation_id,
    )

    return DataFrameNormalizer(
        raw_results=df,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )


@investigations_evidence.command(name="show")
@tracing
def evidence_show(
    evidence_type: Optional[InvestigationEvidenceType] = None,
    database: str = ":memory:",
    investigation_id: Optional[str] = None,
    tenant: Optional[str] = None,
):
    """
    Show currently staged evidence.
    """
    db = find_database(database)
    df = read_database(
        db,
        evidence_type=evidence_type,
        tenant_id=tenant,
        investigation_id=investigation_id,
    )

    return DataFrameNormalizer(
        raw_results=df,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )


@app.command()
@tracing
def create(
    title: Annotated[str, typer.Option(help="Title for the Investigation.")],
    key_findings: Annotated[
        Path, typer.Option(help="Markdown file with key findings.")
    ],
    priority: Annotated[
        InvestigationPriority, typer.Option(help="Investigation Priority.")
    ] = InvestigationPriority.MEDIUM,
    type_: Annotated[
        InvestigationType, typer.Option("--type", help="Investigation Type.")
    ] = InvestigationType.SECURITY_INVESTIGATION,
    status: Annotated[
        InvestigationStatus, typer.Option(help="Investigation Status.")
    ] = InvestigationStatus.OPEN,
    assignee_id: Annotated[
        str,
        typer.Option(
            help="ID for Investigation assignment, may use @me, @partner or @tenant for quick reference."
        ),
    ] = "@tenant",
    database: Annotated[
        str,
        typer.Option(
            help="Investigation Evidence database location.  Can be a file path or ':memory:'."
        ),
    ] = ":memory:",
    dry_run: Annotated[
        bool,
        typer.Option(
            help="Setting to true only prints parameters.  API call is not submitted."
        ),
    ] = False,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Context ID.")] = None,
):
    """
    Create a new Investigation.
    """
    service = get_service(environment=region, tenant_id=tenant)

    alerts = None
    events = None
    search_queries = None

    if database:
        evidence = get_investigation_evidence(database, service.tenant_id, "NEW")
        alerts = evidence.alerts
        events = evidence.events
        search_queries = evidence.search_queries

    # verify and save valid search queries
    if not dry_run:
        if search_queries:
            queries = service.queries.query.ql_queries(
                QLQueriesInput(rns=search_queries)
            )

            search_queries = [query.rn for query in queries.queries or []]
        else:
            search_queries = []

        assignee_id = lookup_assignee_id(service, assignee_id)

    create_investigation_input = CreateInvestigationInput(
        alerts=alerts,
        assignee_id=assignee_id,
        events=events,
        key_findings=key_findings.read_text(),
        priority=INVESTIGATION_PRIORITY_MAP[priority],
        search_queries=search_queries,
        status=status,
        title=title,
        type=type_,
    )

    if dry_run:
        created_investigation = None
    else:
        created_investigation = federated_investigation_create(
            service=service, input_=create_investigation_input
        )

    results = InvestigationsCreatedResultsNormalizer(
        raw_results=create_investigation_input if dry_run else created_investigation,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        dry_run=dry_run,
        arguments=inspect.currentframe().f_locals,
    )

    return results


@investigations_evidence.command(name="append")
@tracing
def evidence_append(
    investigation_id: Annotated[str, typer.Option(help="Investigation Identifier.")],
    database: Annotated[
        str,
        typer.Option(
            help="Investigation Evidence database location.  Can be a file path or ':memory:'."
        ),
    ] = ":memory:",
    use_new: Annotated[
        bool, typer.Option(help="Use NEW investigation id for evidence.")
    ] = False,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Context ID.")] = None,
):
    """
    Append evidence an existing investigation.
    """
    if not database:
        raise ValueError("Database must be provided to append evidence.")

    service = get_service(environment=region, tenant_id=tenant)

    alerts = None
    events = None
    search_queries = None

    evidence = get_investigation_evidence(
        database, service.tenant_id, "NEW" if use_new else investigation_id
    )
    alerts = evidence.alerts
    events = evidence.events
    search_queries = evidence.search_queries

    results = service.investigations2.mutation.add_evidence_to_investigation(
        AddEvidenceToInvestigationInput(
            investigation_id=investigation_id,
            alerts=alerts,
            events=events,
            search_queries=search_queries,
        )
    )
    log.debug("Add evidence API results:", results)

    results = service.investigations2.query.investigation_v2(
        InvestigationV2Arguments(
            id=investigation_id,
        )
    )

    return InvestigationsCreatedResultsNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )


@investigations_evidence.command(name="remove")
@tracing
def evidence_remove(
    investigation_id: Annotated[str, typer.Option(help="Investigation Identifier.")],
    database: Annotated[
        str,
        typer.Option(
            help="Investigation Evidence database location.  Can be a file path or ':memory:'."
        ),
    ] = ":memory:",
    use_new: Annotated[
        bool, typer.Option(help="Use NEW investigation id for evidence.")
    ] = False,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Context ID.")] = None,
):
    """
    Remove evidence an existing investigation.
    """
    if not database:
        raise ValueError("Database must be provided to remove evidence.")

    service = get_service(environment=region, tenant_id=tenant)

    alerts = None
    events = None
    search_queries = None

    evidence = get_investigation_evidence(
        database, service.tenant_id, "NEW" if use_new else investigation_id
    )
    log.debug("Retrieved evidence for investigation:", evidence)
    alerts = evidence.alerts
    events = evidence.events
    search_queries = evidence.search_queries

    results = service.investigations2.mutation.remove_evidence_from_investigation(
        RemoveEvidenceFromInvestigationInput(
            investigation_id=investigation_id,
            alerts=alerts,
            events=events,
            search_queries=search_queries,
        )
    )
    log.debug("Remove evidence API results:", results)

    results = service.investigations2.query.investigation_v2(
        InvestigationV2Arguments(
            id=investigation_id,
        )
    )

    return InvestigationsCreatedResultsNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )


@app.command()
@tracing
def search(
    cell: Optional[str] = None,
    # search_children_tenants: bool = False,
    limit: Optional[int] = None,
    region: Optional[str] = None,
    tenant: Optional[str] = None,
):
    """Taegis Investigations search."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 100

    results = []

    # fix for CX-99036
    pattern = r"\|\s*(head|tail)\s*([0-9]+)"
    match = re.search(pattern, cell)

    if not limit:
        if match and match.group(1) == "tail":  # pragma: no cover
            log.warning(
                "tail is not currently supported, it will be used as the limit..."
            )

        if match:
            limit = int(match.group(2))
    elif match:  # pragma: no cover
        log.warning(
            f"limit and {match.group(1)} both provided, only limit will be honored..."
        )

    cell = re.sub(pattern, "", cell)

    if limit and limit < per_page:
        per_page = limit
    # endfix

    # fix for CX-103490
    output = build_output_string(TaegisMagicInvestigationsV2)

    output = remove_output_node(output, "metric")
    output = remove_output_node(output, "metrics")
    # endfix

    with service(output=output):
        investigations_results = federated_investigations_search(
            service=service,
            arguments=InvestigationsV2Arguments(
                page=page,
                per_page=per_page,
                cql=cell,
            ),
        )

    results.append(investigations_results)

    # fix for CX-99036
    if not limit or investigations_results.total_count < limit:
        limit = investigations_results.total_count
    # endfix

    while (
        sum_results := sum(len(result.investigations) for result in results)
    ) < limit:
        page += 1

        # fix for CX-99036
        if (per_page * page) > limit:
            per_page = limit - sum_results
        # endfix

        with service(output=output):
            investigations_results = federated_investigations_search(
                service=service,
                arguments=InvestigationsV2Arguments(
                    page=page,
                    per_page=per_page,
                    cql=cell,
                ),
            )
        results.append(investigations_results)

    normalized_results = InvestigationsSearchResultsNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@app.command(name="merge")
@tracing
def investigations_merge(
    source: Annotated[str, typer.Option(help="Source Investigation ID.")],
    target: Annotated[str, typer.Option(help="Target Investigation ID.")],
    close_status: Annotated[
        InvestigationStatus,
        typer.Option(
            help="Investigation Status to set for the source investigation.",
        ),
    ] = InvestigationStatus.CLOSED_INFORMATIONAL,
    tenant: Annotated[Optional[str], typer.Option(help="Taegis Tenant ID.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Taegis Region ID.")] = None,
):
    """Merge investigations evidence and close the source investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    source_investigation = service.investigations2.query.investigation_v2(
        InvestigationV2Arguments(
            id=source,
        )
    )

    if not source_investigation:
        raise ValueError("Source Investigation not found.")

    if source_investigation.status not in (
        InvestigationStatus.OPEN,
        InvestigationStatus.DRAFT,
    ):
        raise ValueError("Source Investigation status must be OPEN or DRAFT to merge.")

    if source_investigation.archived_at is not None:
        raise ValueError("Source Investigation must not be archived to merge.")

    if "Merged Investigation" in source_investigation.title:
        raise ValueError("Source Investigation has already been merged.")

    target_investigation = service.investigations2.query.investigation_v2(
        InvestigationV2Arguments(
            id=target,
        )
    )

    if not target_investigation:
        raise ValueError("Target Investigation not found.")

    if target_investigation.status not in (
        InvestigationStatus.OPEN,
        InvestigationStatus.DRAFT,
    ):
        raise ValueError("Target Investigation status must be OPEN or DRAFT to merge.")

    if target_investigation.archived_at is not None:
        raise ValueError("Target Investigation must not be archived to merge.")

    result = service.sharelinks.mutation.create_share_link(
        ShareLinkCreateInput(
            link_ref=source_investigation.id,
            link_type="investigationId",
            tenant_id=tenant,
        )
    )

    source_shareable_url = (
        service.investigations.sync_url.replace("api.", "") + f"/share/{result.id}"
    )

    result = service.sharelinks.mutation.create_share_link(
        ShareLinkCreateInput(
            link_ref=target_investigation.id,
            link_type="investigationId",
            tenant_id=tenant,
        )
    )

    target_shareable_url = (
        service.investigations.sync_url.replace("api.", "") + f"/share/{result.id}"
    )

    results = service.investigations2.mutation.add_comment_to_investigation(
        AddCommentToInvestigationInput(
            investigation_id=source_investigation.id,
            comment=f"Investigation evidence moved into {target_investigation.id} ({target_shareable_url}).",
        )
    )
    log.debug("Add comment to investigation:", results)

    results = service.investigations2.mutation.add_comment_to_investigation(
        AddCommentToInvestigationInput(
            investigation_id=target_investigation.id,
            comment=f"Investigation evidence moved from {source_investigation.id} ({source_shareable_url}).",
        )
    )
    log.debug("Add comment to investigation:", results)

    results = service.investigations2.mutation.update_investigation_v2(
        UpdateInvestigationV2Input(
            id=source_investigation.id,
            status=close_status,
            title=f"Merged Investigation: {source_investigation.title}",
        )
    )
    log.debug("Close source investigation:", results)
    results = service.investigations2.mutation.archive_investigation_v2(
        ArchiveInvestigationInput(id=source_investigation.id)
    )
    log.debug("Archive source investigation:", results)

    results = service.investigations2.mutation.add_evidence_to_investigation(
        AddEvidenceToInvestigationInput(
            investigation_id=target_investigation.id,
            alerts=[
                evidence.alert_id for evidence in source_investigation.alerts_evidence
            ],
            events=[
                evidence.event_id for evidence in source_investigation.events_evidence
            ],
            search_queries=[
                evidence.search_query
                for evidence in source_investigation.search_queries_evidence
            ],
        )
    )
    log.debug("Add evidence to target investigation:", results)

    time.sleep(3)  # allow background workers to process
    target_investigation = service.investigations2.query.investigation_v2(
        InvestigationV2Arguments(
            id=target,
        )
    )

    normalized_results = InvestigationsCreatedResultsNormalizer(
        raw_results=target_investigation,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_search_queries.command(name="add")
@tracing
def investigations_search_queries_add(
    query_id: Annotated[str, typer.Option()],
    tenant_id: Annotated[str, typer.Option()],
    query: Annotated[str, typer.Option()],
    results_returned: Annotated[int, typer.Option()] = 0,
    total_results: Annotated[int, typer.Option()] = 0,
    database: Annotated[str, typer.Option()] = ":memory:",
):
    """Add a Taegis investigations search query."""
    normalized_results = InsertSearchQueryNormalizer(
        query_identifier=query_id,
        tenant_id=tenant_id,
        query=query,
        results_returned=results_returned,
        total_results=total_results,
    )

    insert_search_query(database, normalized_results)

    results = list_search_queries(database)

    normalized_results = DataFrameNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_search_queries.command(name="remove")
@tracing
def investigations_search_queries_remove(
    query_id: Annotated[str, typer.Option()],
    database: Annotated[str, typer.Option()] = ":memory:",
):
    """Remove a Taegis investigations search query."""
    delete_search_query(database, query_id)

    results = list_search_queries(database)

    normalized_results = DataFrameNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_search_queries.command(name="clear")
@tracing
def investigations_search_queries_clear(
    database: Annotated[str, typer.Option()] = ":memory:",
):
    """Remove all Taegis investigations search queries."""
    clear_search_queries(database)

    results = list_search_queries(database)

    normalized_results = DataFrameNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_search_queries.command(name="list")
@tracing
def investigations_search_queries_list(
    database: str = ":memory:",
):
    """List tracked Taegis investigations search queries."""
    results = list_search_queries(database)

    normalized_results = DataFrameNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_search_queries.command(name="stage")
@tracing
def investigations_search_queries_stage(
    database: Annotated[str, typer.Option()] = ":memory:",
    investigation_id: Annotated[str, typer.Option()] = "NEW",
):
    """Stage Taegis investigations search queries to attach to investigation."""
    results = list_search_queries(database)

    db = find_database(database)

    stage_investigation_evidence(
        results, db, InvestigationEvidenceType.Query, investigation_id
    )

    clear_search_queries(database)

    normalized_results = DataFrameNormalizer(
        raw_results=results,
        service="investigations",
        tenant_id="N/A",
        region="N/A",
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_attachment.command(name="list")
@tracing
def investigations_attachment_list(
    investigation_id: Annotated[str, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """List file attachments for a given investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1
    per_page = 20

    files = []

    results = service.investigations2.query.investigation_files_v2(
        InvestigationFilesV2Arguments(
            investigation_id=investigation_id,
            page=page,
            per_page=per_page,
        )
    )
    total_count = results.total_count
    files.extend(results.files)

    remaining_pages = -(-(total_count - per_page) // per_page)

    for page in range(2, remaining_pages + 2):
        results = service.investigations2.query.investigation_files_v2(
            InvestigationFilesV2Arguments(
                investigation_id=investigation_id,
                page=page,
                per_page=per_page,
            )
        )
        files.extend(results.files)

    normalized_results = TaegisResults(
        raw_results=files,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_attachment.command(name="get")
@tracing
def investigations_attachment_get(
    file_id: Annotated[str, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Get a file attachment."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.investigations2.query.investigation_file_v2(
        InvestigationFileV2Arguments(
            file_id=file_id,
        )
    )

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_attachment.command(name="remove")
@tracing
def investigations_attachment_remove(
    file_id: Annotated[str, typer.Option()],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Delete file attachment."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.investigations2.mutation.delete_investigation_file(
        DeleteInvestigationFileInput(
            file_id=file_id,
        )
    )

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_attachment.command(name="upload")
@tracing
def investigations_attachment_upload(
    investigation_id: Annotated[str, typer.Option()],
    file: Annotated[
        Path,
        typer.Option(
            exists=True,
            file_okay=True,
            dir_okay=False,
            writable=False,
            readable=True,
            resolve_path=True,
        ),
    ],
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Upload file attachment."""
    service = get_service(environment=region, tenant_id=tenant)

    file_input = InitInvestigationFileUploadInput(
        investigation_id=investigation_id,
        name=file.name,
        content_type=str(mimetypes.guess_type(file)[0]),
        size=file.stat().st_size,
    )
    log.debug(file_input)

    results = service.investigations2.mutation.init_investigation_file_upload(
        input_=file_input
    )
    log.debug(results)

    with file.open("rb") as f:
        upload_response = requests.put(
            results.presigned_url,
            headers={
                "Accept": "*/*",
                "Content-Type": str(mimetypes.guess_type(file)[0]),
                "Content-Length": str(file.stat().st_size),
            },
            data=f,
        )
    log.debug(upload_response)
    time.sleep(3)

    verify_upload = service.investigations2.query.investigation_file_v2(
        InvestigationFileV2Arguments(
            file_id=results.file.id,
        )
    )
    log.debug(verify_upload)

    normalized_results = TaegisResult(
        raw_results=verify_upload,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_attachment.command("download")
@tracing
def investigations_attachment_download(
    file_id: Annotated[
        str,
        typer.Option(),
    ],
    save_as: Annotated[Optional[str], typer.Option()] = None,
    tenant: Annotated[Optional[str], typer.Option()] = None,
    region: Annotated[Optional[str], typer.Option()] = None,
):
    """Get a file attachment."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.investigations2.query.investigation_file_v2(
        InvestigationFileV2Arguments(
            file_id=file_id,
        )
    )

    if not results.download_url:
        log.error("Cannot download file, no download url found.")
        raise typer.Exit(code=1)

    with requests.get(results.download_url, stream=True) as r:
        r.raise_for_status()

        if save_as:
            filename = save_as
        else:
            filename = results.name

        file_path = verify_file(filename)

        with file_path.open("wb") as f:
            for chunk in r.iter_content(chunk_size=8192):
                f.write(chunk)

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_comment.command(name="list")
@tracing
def investigations_comments_list(
    investigation_id: Annotated[str, typer.Option(help="Investigation Identifier.")],
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Identifier.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
):
    """List comments for an investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    page = 1

    results = service.investigations2.query.comments_v2(
        CommentsV2Arguments(
            investigation_id=investigation_id,
            page=page,
            per_page=100,
        )
    )

    comments = results.comments
    total_count = results.total_count

    while len(comments) < total_count:
        page += 1
        results = service.investigations2.query.comments_v2(
            CommentsV2Arguments(
                investigation_id=investigation_id,
                page=page,
                per_page=100,
            )
        )

        comments.extend(results.comments)

    normalized_results = TaegisResults(
        raw_results=comments,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_comment.command(name="add")
@tracing
def investigations_comments_add(
    investigation_id: Annotated[str, typer.Option(help="Investigation Identifier.")],
    cell: Annotated[str, typer.Option(help="Comment text.")],
    is_internal: Annotated[
        bool, typer.Option(help="Mark comment as internal. For partner use only.")
    ] = False,
    mention: Annotated[
        Optional[List[str]],
        typer.Option(
            help="Mention a user, may use @me, @partner or @tenant for quick reference.  May be used multiple times."
        ),
    ] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Identifier.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
):
    """Add a comment to an investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    for m in mention or []:
        if m.startswith("@"):
            m = lookup_assignee_id(service, m)
        cell += f"\n\n@{m}"

    results = service.investigations2.mutation.add_comment_to_investigation(
        AddCommentToInvestigationInput(
            investigation_id=investigation_id,
            comment=cell,
            is_internal=is_internal,
        )
    )

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_comment.command(name="update")
@tracing
def investigations_comments_update(
    comment_id: Annotated[str, typer.Option(help="Comment Identifier.")],
    cell: Annotated[str, typer.Option(help="Comment text.")],
    mark_as_read: Annotated[
        bool, typer.Option(help="Mark comment as read. For partner use only.")
    ] = False,
    mention: Annotated[
        Optional[List[str]],
        typer.Option(
            help="Mention a user, may use @me, @partner or @tenant for quick reference.  May be used multiple times."
        ),
    ] = None,
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Identifier.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
):
    """Update a comment to an investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    for m in mention or []:
        if m.startswith("@"):
            m = lookup_assignee_id(service, m)
        cell += f"\n{m}"

    results = service.investigations2.mutation.update_investigation_comment(
        UpdateInvestigationCommentInput(
            comment_id=comment_id,
            comment=cell,
            mark_as_read=mark_as_read,
        )
    )

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results


@investigations_comment.command(name="remove")
@tracing
def investigations_comments_remove(
    comment_id: Annotated[str, typer.Option(help="Comment Identifier.")],
    tenant: Annotated[Optional[str], typer.Option(help="Tenant Identifier.")] = None,
    region: Annotated[Optional[str], typer.Option(help="Region Identifier.")] = None,
):
    """Remove a comment to an investigation."""
    service = get_service(environment=region, tenant_id=tenant)

    results = service.investigations2.mutation.delete_investigation_comment(
        DeleteInvestigationCommentInput(
            comment_id=comment_id,
        )
    )

    normalized_results = TaegisResult(
        raw_results=results,
        service="investigations",
        tenant_id=service.tenant_id,
        region=service.environment,
        arguments=inspect.currentframe().f_locals,
    )

    return normalized_results
